# coding: utf-8
# 2020/1/10 @ tongshiwei

import pathlib
import warnings
from heapq import nlargest
from longling.ML.toolkit.analyser import get_max, get_min, get_by_key, key_parser
from longling import as_list, path_append
from longling import list2dict, nested_update

import json
import sqlite3
import os


def _key(x):
    """
    Examples
    --------
    >>> _key(123)
    123.0
    >>> _key('{"default": 123}')
    123.0
    """
    try:
        return float(x)
    except ValueError:
        return float(json.loads(x)["default"])


def show(key, max_key=True, exp_id=None, res_dir="./",
         nni_dir=path_append(os.environ.get("HOME", "./"), "nni-experiments"),
         only_final=False,
         with_keys=None, with_all=False):  # pragma: no cover
    """
    Updated in v1.3.17

    cli alias: ``nni show``

    Parameters
    ----------
    key
    exp_id
    res_dir
    nni_dir
    only_final
    with_keys
    with_all

    Returns
    -------

    """
    if exp_id is None:
        exp_id = pathlib.PurePath(os.path.abspath(res_dir)).name
    nni_dir = path_append(nni_dir, exp_id)
    sqlite_db = path_append(nni_dir, "db", "nni.sqlite", to_str=True)
    print(sqlite_db)
    conn = sqlite3.connect(sqlite_db)
    c = conn.cursor()
    if only_final:
        cursor = c.execute("SELECT trialJobId FROM MetricData WHERE type='FINAL';")
    else:
        cursor = c.execute("SELECT DISTINCT trialJobId FROM MetricData;")
    trial_dir = path_append(nni_dir, "trials")
    result = []
    for trial in [row[0] for row in cursor]:
        with open(path_append(trial_dir, trial, "parameter.cfg")) as f:
            trial_params = json.load(f)["parameters"]
        trial_res = path_append(res_dir, trial, "result.json", to_str=True)
        if max_key:
            value, appendix = get_max(
                trial_res, key, with_keys=with_keys, with_all=with_all, merge=False
            )
        else:
            value, appendix = get_min(
                trial_res, key, with_keys=with_keys, with_all=with_all, merge=False
            )
        if with_keys is not None or with_all is True:
            result.append([trial, trial_params, value, dict(appendix[key])])
        else:
            result.append([trial, trial_params, value])
    conn.close()
    result.sort(key=lambda x: float(x[2][key]), reverse=True)
    return result


def show_top_k(k, exp_id=None,
               exp_dir=path_append(os.environ.get("HOME", "./"), "nni-experiments")):  # pragma: no cover
    """
    Updated in v1.3.17

    cli alias: ``nni k-best``

    Parameters
    ----------
    k
    exp_id
    exp_dir

    Returns
    -------

    """
    import warnings
    warnings.warn("deprecated method")

    if exp_id:
        exp_dir = path_append(exp_dir, exp_id)
    sqlite_db = path_append(exp_dir, "db", "nni.sqlite", to_str=True)
    print(sqlite_db)
    conn = sqlite3.connect(sqlite_db)
    c = conn.cursor()
    cursor = c.execute("SELECT trialJobId, data FROM MetricData WHERE type='FINAL';")
    _ret = []
    top_k = nlargest(k, [row for row in cursor], key=lambda x: _key(x[1]))
    trial_dir = path_append(exp_dir, "trials")
    for trial, result in sorted(top_k, key=lambda x: _key(x[1]), reverse=True):
        with open(path_append(trial_dir, trial, "parameter.cfg")) as f:
            trial_params = json.load(f)["parameters"]
            _ret.append([trial, result, trial_params])
    conn.close()

    return _ret


class BaseReporter(object):
    def intermediate(self, data):
        raise NotImplementedError

    def final(self):
        raise NotImplementedError


def get_params(received_params: dict):
    """
    Updated in v1.3.17

    Parameters
    ----------
    received_params: dict
        nni get_next_parameters() 得到的参数字典

    Returns
    -------
    cfg_params: dict
        更新后的参数字典

    Examples
    --------
    >>> get_params({
    ...     "hyper_params_update:hidden_num": 50,
    ...     "hyper_params_update:alpha": 5,
    ...     "learning_rate": 0.1
    ... })
    {'hyper_params_update': {'hidden_num': 50, 'alpha': 5}, 'learning_rate': 0.1}
    """
    cfg_params = {}

    for k, v in received_params.items():
        if ":" in k:
            k = k.split(":")
            obj = list2dict(k, v)
        else:
            obj = {k: v}

        nested_update(cfg_params, obj)

    return cfg_params


def prepare_hyper_search(cfg_kwargs: dict,
                         reporthook=None, final_reporthook=None,
                         primary_key=None, max_key=True, reporter_cls=None, with_keys: (list, str, None) = None,
                         final_keys: (list, str, None) = None,
                         dump=False, disable=False):
    """
    Updated in v1.3.18

    从 nni package 中获取超参，更新配置文件参数。当 nni 不可用或不是 nni 搜索模式时，参数将不会改变。

    .. code-block :: python

        cfg_kwargs, reporthook, final_reporthook, tag = prepare_hyper_search(
            cfg_kwargs, Configuration, reporthook, final_reporthook, primary_key="macro_avg:f1"
        )

        _cfg = Configuration(**cfg_kwargs)
        model = Model(_cfg)
        ...

        for epoch in range(_cfg.begin_epoch, _cfg.end_epoch):
            for batch_data in dataset:
                train_model(batch_data)

            data = evaluate_model()
            reporthook(data)

        final_reporthook()

    Parameters
    ----------
    cfg_kwargs: dict
        待传入cfg的参数
    reporthook
    final_reporthook
    primary_key:
        评估模型用的主键,
        ``nni.report_intermediate_result`` 和 ``nni.report_final_result`` 中  ``metric`` 的 ``default``
    max_key: bool
        主键是越大越好
    reporter_cls
    with_keys: list or str
        其它要存储的 metric，final report时默认为 primary_key 最优时指标
    final_keys: list or str
        with_keys 中使用最后一个 report result 而不是 primary_key 最优时指标
    dump: bool
        为 True 时，会修改 配置文件 中 workspace 参数为 ``workspace/nni.get_experiment_id()/nni.get_trial_id()``
        使得 nni 的中间结果会被存储下来。
    disable

    Returns
    -------
    cfg_kwargs: dict
        插入了nni超参后的配置文件参数
    reporthook: function
        每个iteration结束后的回调函数，用来报告中间结果。
        默认 ``nni.report_intermediate_result``。
    final_reporthook:
        所有iteration结束后的回调函数，用来报告最终结果。
        默认 ``nni.report_final_result``
    dump: bool
        和传入参数保持一致

    Examples
    --------
    .. code-block :: python

        class CFG(Configuration):
            hyper_params = {"hidden_num": 100}
            learning_rate = 0.001
            workspace = ""

        cfg_kwargs, reporthook, final_reporthook, dump = prepare_hyper_search(
            {"learning_rate": 0.1}, CFG, primary_key="macro_avg:f1", with_keys="accuracy"
        )
        # cfg_kwargs: {'learning_rate': 0.1}

    when nni start (e.g., using ``nni create --config _config.yml``),
    suppose in ``_config.yml``:

    .. code-block: yml

        searchSpacePath: _search_space.json

    and in ``_search_space.json``

    .. code-block :: json

        {
            "hidden_num": {"_type": "choice", "_value": [500, 600, 700, 835, 900]},
        }

    one of the return cfg_kwargs is ``{'hyper_params': {'hidden_num': 50}, 'learning_rate': 0.1}``
    """
    if disable:
        return cfg_kwargs, None, None, None
    try:
        import nni
        from nni import get_next_parameter, report_intermediate_result, report_final_result

        assert primary_key is not None

        def _as_key_list(_keys: (list, str, None)):
            if isinstance(_keys, str):
                if ";" in _keys:
                    _keys = _keys.split(";")
                else:
                    _keys = [_keys]
            elif isinstance(_keys, list):
                pass
            elif _keys is None:
                _keys = []
            return _keys

        with_keys = _as_key_list(with_keys)
        final_keys = _as_key_list(final_keys)

        class Reporter(BaseReporter):
            def __init__(self):
                self.datas = []

            def intermediate(self, data):
                feed_dict = {
                    'default': float(get_by_key(data, key_parser(primary_key))),
                    primary_key: get_by_key(data, key_parser(primary_key))
                }
                for key in with_keys:
                    feed_dict[key] = get_by_key(data, key_parser(key))
                report_intermediate_result(feed_dict)
                self.datas.append(data)

            def final(self):
                best_fn = get_min if max_key is False else get_max
                _with_keys = (with_keys if with_keys else []) + [primary_key]
                _final_keys = set(final_keys if final_keys else [])
                final_result = best_fn(
                    self.datas, primary_key, with_keys=";".join(_with_keys), merge=False
                )
                feed_dict = {
                    'default': float(final_result[0][primary_key])
                }
                appendix_dict = dict(final_result[1][primary_key])
                for key in _with_keys:
                    if key in _final_keys:
                        feed_dict[key] = get_by_key(self.datas[-1], key_parser(key))
                    else:
                        feed_dict[key] = appendix_dict[key]
                report_final_result(feed_dict)

        rc = Reporter() if reporter_cls is None else reporter_cls
        reporthook = reporthook if reporthook is not None else rc.intermediate
        final_reporthook = final_reporthook if final_reporthook is not None else rc.final
        cfg_cls_params = get_params(get_next_parameter())
        using_nni_tag = True if cfg_cls_params else False
        nested_update(cfg_kwargs, cfg_cls_params)
        if using_nni_tag is True and dump is True:  # pragma: no cover
            cfg_kwargs["workspace"] = cfg_kwargs.get("workspace", "") + path_append(
                nni.get_experiment_id(), nni.get_trial_id(), to_str=True
            )
        return cfg_kwargs, reporthook, final_reporthook, dump

    except ModuleNotFoundError:  # pragma: no cover
        warnings.warn("nni package not found, skip")
        return cfg_kwargs, reporthook, final_reporthook, dump



# coding: utf-8
# 2021/2/12 @ tongshiwei

from longling.ML.DL import light_module as dlm
from .glue import net_init as _net_init
from .toolkit import loss_dict2tmt_mx_loss, get_trainer as _get_trainer, save_params as _save_params


def train(
        net, cfg, loss_function, trainer,
        train_data, test_data=None,
        params_save=False, dump_result=False, progress_monitor=None,
        *,
        fit_f, eval_f=None, get_net=None, get_loss=None, get_trainer=None, save_params=None,
        enable_hyper_search=False, reporthook=None, final_reporthook=None, primary_key=None,
        eval_epoch=1, initial_net=True, net_init=None, verbose=None, dump_cfg=None,
        **cfg_kwargs
):
    if initial_net:
        net_init = _net_init if net_init is None else net_init
    else:
        net_init = None

    if trainer is None and get_trainer is None:
        get_trainer = _get_trainer

    save_params = _save_params if save_params is None else save_params

    dlm.train(
        net, cfg, loss_function, trainer, train_data, test_data, params_save, dump_result, progress_monitor,
        fit_f=fit_f, eval_f=eval_f,
        net_init=net_init, get_net=get_net, get_loss=get_loss, get_trainer=get_trainer,
        save_params=save_params,
        enable_hyper_search=enable_hyper_search, reporthook=reporthook,
        final_reporthook=final_reporthook,
        primary_key=primary_key, eval_epoch=eval_epoch, loss_dict2tmt_loss=loss_dict2tmt_mx_loss, verbose=verbose,
        dump_cfg=dump_cfg,
        **cfg_kwargs
    )



# coding: utf-8
# create by tongshiwei on 2018/7/14

# from .ValueMonitor import ValueMonitor, EMAValue, as_tmt_value
from longling import as_list
from longling.ML.toolkit.monitor.ValueMonitor import ValueMonitor, EMAValue, as_tmt_value
__all__ = ["LossMonitor", "MovingLoss", "as_tmt_loss", "loss_dict2tmt_loss"]


# def tmt_loss(loss2value=lambda x: x):
#     return tmt_value(transform=loss2value)


def as_tmt_loss(loss_obj, loss2value=lambda x: x):
    """

    Parameters
    ----------
    loss_obj
    loss2value

    Returns
    -------

    Examples
    --------
    >>> @as_tmt_loss
    ... def mse(v):
    ...     return v ** 2
    >>> mse(2)
    4
    """
    return as_tmt_value(loss_obj, loss2value)


def loss_dict2tmt_loss(loss_dict, loss2value=lambda x: x, exclude=None, include=None, as_loss=as_tmt_loss):
    """

    Parameters
    ----------
    loss_dict
    loss2value
    exclude
    include
    as_loss

    Returns
    -------

    Examples
    --------
    >>> def mse(v):
    ...     return v ** 2
    >>> losses = loss_dict2tmt_loss({"mse": mse, "rmse": lambda x: x})
    >>> losses.keys()
    dict_keys(['mse', 'rmse'])
    >>> ema = EMAValue(losses)
    >>> losses["mse"](2)
    4
    >>> losses["rmse"](2)
    2
    >>> ema.items()
    dict_items([('mse', 4), ('rmse', 2)])
    >>> losses = loss_dict2tmt_loss({"mse": mse, "rmse": lambda x: x}, include="mse")
    >>> losses.keys()
    dict_keys(['mse', 'rmse'])
    >>> ema = EMAValue(losses, auto="ignore")
    >>> losses["mse"](2)
    4
    >>> losses["rmse"](2)
    2
    >>> ema.items()
    dict_items([('mse', 4), ('rmse', nan)])
    >>> losses = loss_dict2tmt_loss({"mse": mse, "rmse": lambda x: x}, exclude="mse")
    >>> losses.keys()
    dict_keys(['mse', 'rmse'])
    >>> ema = EMAValue(losses, auto="ignore")
    >>> losses["mse"](2)
    4
    >>> losses["rmse"](2)
    2
    >>> ema.items()
    dict_items([('mse', nan), ('rmse', 2)])
    """
    exclude = set() if exclude is None else set(as_list(exclude))
    if include is not None:
        include = set(as_list(include))
        return {
            name: as_loss(func, loss2value) if name in include else func for name, func in loss_dict.items()
        }
    return {
        name: as_loss(func, loss2value) if name not in exclude else func for name, func in loss_dict.items()
    }


class LossMonitor(ValueMonitor):
    @property
    def losses(self):
        return self.value


class MovingLoss(EMAValue, LossMonitor):
    """
    Examples
    --------
    >>> lm = MovingLoss(["l2"])
    >>> lm.losses
    {'l2': nan}
    >>> lm("l2", 100)
    >>> lm("l2", 1)
    >>> lm["l2"]
    90.1
    """
    pass



# coding: utf-8
# Copyright @tongshiwei
from __future__ import absolute_import

from .configuration import Configuration, ConfigurationParser
from .etl import transform, etl
from .module import Module
from .sym import get_bp_loss, fit_f, net_init

__all__ = [
    "Module",
    "Configuration", "ConfigurationParser",
    "get_bp_loss",
    "etl", "transform", "fit_f", "net_init"
]



# coding: utf-8
# 2021/3/15 @ tongshiwei

from copy import deepcopy
from longling.ML.DL.optimizer_cfg import OPTIMIZERS as _OPTIMIZERS

OPTIMIZERS = deepcopy(_OPTIMIZERS)

for optimizer in OPTIMIZERS.values():
    params = optimizer[1]
    params["lr"] = params.pop("learning_rate")
    params["weight_decay"] = params.pop("wd")
    params.pop("clip_gradient")


def get_optimizer_cfg(name):
    """
    Examples
    -------
    >>> get_optimizer_cfg('base')
    ('Adam', {'lr': 0.001, 'weight_decay': 0.0001})
    """
    try:
        optimizer, optimizer_params = OPTIMIZERS[name]
    except KeyError:
        raise KeyError("the name should be in: %s" % ", ".join(OPTIMIZERS))
    optimizer_params = deepcopy(optimizer_params)

    return optimizer, optimizer_params



# coding: utf-8

"""日志设定文件"""

import logging

from longling.lib.stream import build_dir
from longling.lib.time import get_current_timestamp as default_timestamp

__all__ = ["LogLevel", "config_logging", "default_timestamp", "set_logging_info"]


class LogLevel(object):
    ERROR = logging.ERROR
    WARN = logging.WARN
    INFO = logging.INFO
    DEBUG = logging.DEBUG
    CRITICAL = logging.CRITICAL


LogLevelDict = {}
for key, value in {
    "error": LogLevel.ERROR,
    "warn": LogLevel.WARN,
    "info": LogLevel.INFO,
    "debug": LogLevel.DEBUG,
    "critical": LogLevel.CRITICAL,
}.items():
    LogLevelDict[key] = value
    LogLevelDict[key.upper()] = value

LOG_COLORS = {
    'WARNING': 'y',
    'INFO': 'g',
    'DEBUG': 'b',
    'CRITICAL': 'y',
    'ERROR': 'r'
}


class _ColoredFormatter(logging.Formatter):
    """
    Formatter for colored log.
    Source code can be found in https://github.com/yxonic/fret/blob/master/fret/__init__.py
    """

    def format(self, record):
        levelname = record.levelname
        if levelname in LOG_COLORS:
            record.levelname = _colored(
                record.levelname,
                LOG_COLORS[record.levelname],
                style='b'
            )
        return logging.Formatter.format(self, record)


def _colored(fmt, fg=None, bg=None, style=None):
    """
    Return colored string.

    List of colours (for fg and bg):
        - k:   black
        - r:   red
        - g:   green
        - y:   yellow
        - b:   blue
        - m:   magenta
        - c:   cyan
        - w:   white

    List of styles:
        - b:   bold
        - i:   italic
        - u:   underline
        - s:   strike through
        - x:   blinking
        - r:   reverse
        - y:   fast blinking
        - f:   faint
        - h:   hide

    Parameters
    ----------
    fmt : str
        string to be colored
    fg : str
        foreground color
    bg : str
        background color
    style : str
        text style

    Examples
    --------
    >>> _colored("info")
    'info'
    >>> ass_str = r'\x1b[34;41minfo\x1b[0m'
    >>> ass_str == _colored("info", "b", "r")
    True
    """

    colcode = {
        'k': 0,  # black
        'r': 1,  # red
        'g': 2,  # green
        'y': 3,  # yellow
        'b': 4,  # blue
        'm': 5,  # magenta
        'c': 6,  # cyan
        'w': 7  # white
    }

    fmtcode = {
        'b': 1,  # bold
        'f': 2,  # faint
        'i': 3,  # italic
        'u': 4,  # underline
        'x': 5,  # blinking
        'y': 6,  # fast blinking
        'r': 7,  # reverse
        'h': 8,  # hide
        's': 9,  # strike through
    }

    # properties
    props = []
    if isinstance(style, str):
        props = [fmtcode[s] for s in style]
    if isinstance(fg, str):
        props.append(30 + colcode[fg])
    if isinstance(bg, str):
        props.append(40 + colcode[bg])

    # display
    props = ';'.join([str(x) for x in props])
    if props:
        return '\x1b[%sm%s\x1b[0m' % (props, fmt)
    else:
        return fmt


def config_logging(filename=None,
                   log_format=None,
                   level=logging.INFO,
                   logger=None, console_log_level=None, propagate=False,
                   mode='a',
                   file_format=None, encoding: (str, None) = "utf-8",
                   enable_colored=False, datefmt=None):
    """
    主日志设定文件

    Parameters
    ----------
    filename: str or None
        日志存储文件名，不为空时将创建文件存储日志
    log_format: str
        默认日志输出格式: %(name)s, %(levelname)s %(message)s
        如果 datefmt 被指定，则变为 %(name)s: %(asctime)s, %(levelname)s %(message)s
    level: str or int
        默认日志等级
    logger: str or logging.logger
        日志logger名，可以为空（使用root logger），
        字符串类型（创建对应名logger），logger
    console_log_level: str, int or None
        屏幕日志等级，不为空时，使能屏幕日志输出
    propagate: bool
    mode: str
    file_format: str or None
        文件日志输出格式，为空时，使用log_format
    encoding
    enable_colored: bool
    datefmt: str

    Returns
    -------

    """
    if logger is None:
        logger = logging.getLogger()
    elif isinstance(logger, str):
        logger = logging.getLogger(logger)

    if isinstance(level, str):
        level = LogLevelDict[level]
    if isinstance(console_log_level, str):
        console_log_level = LogLevelDict[console_log_level]

    # need a clean state, for some module may
    # have called logging functions already (i.e. logging.info)
    # in that case, a default handler would been appended,
    # causing undesired output to stderr
    for handler in logger.handlers:
        logger.removeHandler(handler)

    formatter_cls = logging.Formatter if enable_colored is False else _ColoredFormatter

    if datefmt is None:
        log_format = '%(name)s, %(levelname)s %(message)s' if log_format is None else log_format
    else:
        log_format = '%(name)s: %(asctime)s, %(levelname)s %(message)s' if log_format is None else log_format
        if datefmt == "default":
            datefmt = "%Y-%m-%d %H:%M:%S"

    formatter = formatter_cls(log_format, datefmt=datefmt)
    logger.setLevel(level)
    if not propagate:
        logger.propagate = False
    if filename and filename is not None:
        build_dir(filename)
        handler = logging.FileHandler(filename, mode=mode, encoding=encoding)
        file_formatter = formatter
        if file_format:
            file_formatter = formatter_cls(file_format, datefmt=datefmt)
        handler.setFormatter(file_formatter)
        logger.addHandler(handler)
    if console_log_level is not None:
        ch = logging.StreamHandler()
        ch.setFormatter(formatter_cls(log_format, datefmt=datefmt))
        ch.setLevel(console_log_level)
        logger.addHandler(ch)

    return logger


def set_logging_info():
    logging.getLogger().setLevel(logging.INFO)



# coding: utf-8
# 2021/3/14 @ tongshiwei


import logging
from longling.ML.PytorchHelper.toolkit import net_initialize, load_net


def net_init(
        net, cfg=None,
        force_init=False,
        allow_reinit=True, logger=logging, initialized=False, model_file=None,
        initializer_kwargs=None, *args, **kwargs
):
    if initialized and not force_init:
        logger.warning("model has been initialized, skip model_init")

    try:
        if model_file is not None:
            net = load_net(model_file, cfg.ctx)
            logger.info(
                "load params from existing model file "
                "%s" % model_file
            )
        else:
            raise FileExistsError()
    except FileExistsError:
        if allow_reinit:
            logger.info("model doesn't exist, initializing")
            initializer_kwargs = {} if initializer_kwargs is None else initializer_kwargs
            net_initialize(net, cfg.ctx, **initializer_kwargs)
        else:
            logger.error(
                "model doesn't exist, target file: %s" % model_file
            )



# coding: utf-8
# 2019/12/9 @ tongshiwei
import os
import gzip
import shutil
import tarfile
import zipfile
import rarfile

from longling import flush_print
from longling.lib.candylib import format_byte_sizeof

__all__ = ["decompress", "get_path", "un_zip", "un_rar", "un_tar", "reporthook4urlretrieve"]


def decompress(file):  # pragma: no cover
    for z in [".tar.gz", ".tar.bz2", ".tar.bz", ".tar.tgz", ".tar", ".tgz", ".zip", ".rar", ".gz"]:
        if file.endswith(z):
            if z == ".zip":
                un_zip(file)
            elif z == ".rar":
                un_rar(file)
            elif z in {".tar.bz2", ".tar.bz", ".tar.tgz", ".tar", ".tgz"}:
                un_tar(file)
            elif z == ".gz":
                un_gzip(file)
            break


def get_path(file):  # pragma: no cover
    #  返回解压缩后的文件名
    for i in [".tar.gz", ".tar.bz2", ".tar.bz", ".tar.tgz", ".tar", ".tgz", ".zip", ".rar", ".gz"]:
        file = file.replace(i, "")
    return file


def un_zip(file):  # pragma: no cover
    zip_file = zipfile.ZipFile(file)
    uz_path = get_path(file)
    print(file + " is unzip to " + uz_path)
    for name in zip_file.namelist():
        zip_file.extract(name, uz_path)
    zip_file.close()


def un_rar(file):  # pragma: no cover
    rar_file = rarfile.RarFile(file)
    uz_path = get_path(file)
    print(file + " is unrar to " + uz_path)
    rar_file.extractall(uz_path)


def un_tar(file):  # pragma: no cover
    tar_file = tarfile.open(file)
    uz_path = get_path(file)
    print(file + " is untar to " + uz_path)
    tar_file.extractall(path=uz_path)


def un_gzip(file):  # pragma: no cover
    uz_file = get_path(file)
    with gzip.open(file, 'rb') as f_in:
        with open(uz_file, 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)
    os.remove(file)


def reporthook4urlretrieve(blocknum, bs, size):
    """

    Parameters
    ----------
    blocknum:
        已经下载的数据块
    bs:
        数据块的大小
    size:
        远程文件的大小

    Returns
    -------

    """
    per = 100.0 * (blocknum * bs) / size
    if per > 100:
        per = 100
    flush_print(
        'Downloading %.2f%% : %s | %s' % (
            per,
            format_byte_sizeof(blocknum * bs),
            format_byte_sizeof(size)
        ))



# coding:utf-8
# created by tongshiwei on 2018/8/8
from mxnet import Context, cpu, gpu

from longling.lib.candylib import as_list
from longling.lib.parser import get_parsable_var, var2exp, path_append, Configuration as Params, ConfigurationParser

__all__ = [
    "MXCtx", "Configuration",
    "var2exp", "eval_var",
    "path_append", "ConfigurationParser",
    "cpu", "gpu"
]


def eval_var(variable):
    return eval(variable)


class MXCtx(object):
    @staticmethod
    def load(data):
        ctx_vars = []
        for device_type, device_ids in data.items():
            if isinstance(device_ids, int):
                device_ids = as_list(device_ids)
            elif isinstance(device_ids, str):
                device_ids = map(int, device_ids.split(','))
            for device_id in device_ids:
                ctx_vars.append(eval(device_type)(device_id))
        return ctx_vars

    @staticmethod
    def dump(data):
        ctx_vars = {}
        for ctx in as_list(data):
            assert isinstance(ctx, Context)
            if ctx.device_type not in ctx_vars:
                ctx_vars[ctx.device_type] = []
            ctx_vars[ctx.device_type].append(ctx.device_id)
        for device_type, device_ids in ctx_vars.items():
            if len(device_ids) > 1:
                ctx_vars[device_type] = ",".join(list(map(str, device_ids)))
            else:
                ctx_vars[device_type] = device_ids[0]
        return ctx_vars


class Configuration(Params):
    @property
    def parsable_var(self):
        """
        获取可以进行命令行设定的参数

        Returns
        -------
        store_vars: dict
            可以进行命令行设定的参数
        """
        return get_parsable_var(
            self,
            parse_exclude=None,
            dump_parse_functions={'ctx': MXCtx.dump}
        )

    @classmethod
    def load_parse_function(cls):
        return {"ctx": MXCtx.load}


if __name__ == '__main__':
    print(eval_var(
        "[gpu(0), gpu(1), gpu(2)]"
    ))



# coding: utf-8
# create by tongshiwei on 2019/4/7

from .ValueMonitor import EMAValue, as_tmt_value
from .LossMonitor import MovingLoss, as_tmt_loss
from .ProgressMonitor import ConsoleProgressMonitor



# coding: utf-8
# 2021/3/13 @ tongshiwei

import torch
import os
from .parallel import set_device


def net_initialize(net, model_ctx, *args, **kwargs):
    return set_device(net, model_ctx, *args, **kwargs)


def load_net(filename, net, allow_missing=False, *args, **kwargs):
    """
        Load the existing net parameters

        Parameters
        ----------
        filename: str
            The model file
        net: torch.nn.Module
            The network which has been initialized or
            loaded from the existed model
        allow_missing: bool

        Returns
        -------
        The initialized net
        """
    # 根据文件名装载已有的网络参数
    if not os.path.isfile(filename):
        raise FileExistsError("%s does not exist" % filename)
    net.load_state_dict(torch.load(filename), strict=not allow_missing)
    return net



# coding: utf-8
# 2019/9/20 @ tongshiwei

__all__ = ["cli"]

import os

from longling.Architecture.cli.units import *
from longling.Architecture.install_proj import project_types
from longling.Architecture.utils import legal_input, binary_legal_input, default_legal_input
from longling.lib.path import path_append

from .. import config


def cli(skip_top=True, project=None, override=None, tar_dir="./", **kwargs):
    """
    The main function for arch
    """
    config.OVERRIDE = override

    main_params = main_cli(skip_top, project, **kwargs)

    kwargs.update(main_params)

    indicator = dict(
        travis=binary_legal_input("Install travis?"),
    )

    docs = binary_legal_input("Install docs?")
    if main_params["project_type"] == "docs" or docs:
        default_style = "mxnet" if main_params["project_type"] == "docs" else "sphinx"
        docs_params = docs_cli(default_style=default_style, **kwargs)
    else:
        docs_params = {}

    service_params = {}
    docker_params = {}
    if main_params["project_type"] != "docs":
        service = binary_legal_input("To deploy as a service")
        if service:
            docker_params.update(
                dockerfile_cli(project_type=main_params["project_type"])
            )
            if "port" not in docker_params:
                port = default_legal_input("Image Port", __default_value=None)
            else:
                port = docker_params["port"]
            service_params["port"] = None if port == 'null' else port
            service_params.update(dict(
                private=binary_legal_input("Is private project?"),
            ))
            stages_candidates = {
                "test": {"stage_image_name": docker_params["image_name"]}
            }
            if main_params["project_type"] == "python":
                stages_candidates["build"] = {"need": "n"}
            if binary_legal_input("Install .gitlab-ci.yml?"):
                service_params.update(dict(
                    gitlab_ci_params=gitlab_ci_cli(
                        port=port, docs=docs,
                        stages_candidates=stages_candidates
                    ),
                ))

        elif binary_legal_input("Install Dockerfile?"):
            docker_params.update(dockerfile_cli(project_type=main_params["project_type"]))

    if skip_top:
        tar_dir = path_append(tar_dir, "./")
    else:
        tar_dir = path_append(tar_dir, main_params["project"])
        os.makedirs(tar_dir)

    __project_type = main_params["project_type"]

    project_types[__project_type](
        tar_dir=tar_dir, main_params=main_params, docs_params=docs_params,
        docker_params=docker_params, service_params=service_params, **indicator
    )


if __name__ == '__main__':
    cli()



# coding: utf-8
# created by tongshiwei on 17-12-17

from __future__ import absolute_import

from longling.base import *

from longling.lib.stream import *
from longling.lib.clock import *
from longling.lib.utilog import *
from longling.lib.path import *
from longling.lib.parser import *
from longling.lib.candylib import *
from longling.lib.loading import *
from longling.lib.iterator import *
from longling.lib.concurrency import *
from longling.lib.structure.dict import nested_update

from longling.lib import clock
from longling.lib import regex
from longling.lib import stream
from longling.lib import utilog
from longling.lib import candylib
from longling.lib import path
from longling.lib import parser

__version__ = '1.3.32'



# coding: utf-8
# 2019/12/20 @ tongshiwei

from collections import OrderedDict
from longling.lib.candylib import as_list
from longling import PATH_TYPE, loading

__all__ = ["get_max", "get_by_key", "get_best", "get_min"]

from .utils import get_by_key


def get_best(src: (PATH_TYPE, list), *keys, with_keys: (str, None) = None, with_all=False, cmp=lambda x, y: x > y,
             merge=True):
    keys = as_list(keys)

    with_keys = [] if with_keys is None else with_keys.split(";")

    result = {
        key: None for key in keys
    }

    result_appendix = {
        key: None for key in keys
    }

    for data in loading(src):
        for key in result:
            _data = get_by_key(data, parsed_key=key)
            if result[key] is None or cmp(_data, result[key]):
                result[key] = _data
                if with_all:
                    result_appendix[key] = data
                elif with_keys:
                    result_appendix[key] = {_key: get_by_key(data, _key) for _key in with_keys}

    if merge:
        return _merge(result, result_appendix if with_all or with_keys else None)
    else:
        return result, result_appendix if with_all or with_keys else None


def get_max(src: (PATH_TYPE, list), *keys, with_keys: (str, None) = None, with_all=False, merge=True):
    """
    Examples
    -------
    >>> src = [
    ... {"Epoch": 0, "macro avg": {"f1": 0.7}, "loss": 0.04, "accuracy": 0.7},
    ... {"Epoch": 1, "macro avg": {"f1": 0.88}, "loss": 0.03, "accuracy": 0.8},
    ... {"Epoch": 1, "macro avg": {"f1": 0.7}, "loss": 0.02, "accuracy": 0.66}
    ... ]
    >>> result, _ = get_max(src, "accuracy", merge=False)
    >>> result
    {'accuracy': 0.8}
    >>> _, result_appendix = get_max(src, "accuracy", with_all=True, merge=False)
    >>> result_appendix
    {'accuracy': {'Epoch': 1, 'macro avg': {'f1': 0.88}, 'loss': 0.03, 'accuracy': 0.8}}
    >>> result, result_appendix = get_max(src, "accuracy", "macro avg:f1", with_keys="Epoch", merge=False)
    >>> result
    {'accuracy': 0.8, 'macro avg:f1': 0.88}
    >>> result_appendix
    {'accuracy': {'Epoch': 1}, 'macro avg:f1': {'Epoch': 1}}
    >>> get_max(src, "accuracy", "macro avg:f1", with_keys="Epoch")
    {'accuracy': {'Epoch': 1, 'accuracy': 0.8}, 'macro avg:f1': {'Epoch': 1, 'macro avg:f1': 0.88}}
    """
    return get_best(
        src, *keys, with_keys=with_keys, with_all=with_all, merge=merge
    )


def get_min(src: (PATH_TYPE, list), *keys, with_keys: (str, None) = None, with_all=False, merge=True):
    """
    >>> src = [
    ... {"Epoch": 0, "macro avg": {"f1": 0.7}, "loss": 0.04, "accuracy": 0.7},
    ... {"Epoch": 1, "macro avg": {"f1": 0.88}, "loss": 0.03, "accuracy": 0.8},
    ... {"Epoch": 1, "macro avg": {"f1": 0.7}, "loss": 0.02, "accuracy": 0.66}
    ... ]
    >>> get_min(src, "loss")
    {'loss': 0.02}
    """
    return get_best(
        src, *keys, with_keys=with_keys, with_all=with_all, cmp=lambda x, y: x < y, merge=merge
    )


def _merge(result: dict, appendix):
    """
    Examples
    --------
    >>> _merge({"accuracy": 0.7}, {"accuracy": {"Epoch": 1, "accuracy": 0.7}})
    {'accuracy': {'Epoch': 1, 'accuracy': 0.7}}
    """
    if appendix:
        for key in appendix:
            appendix[key][key] = result[key]
        return appendix
    else:
        return result


if __name__ == '__main__':
    src = [
        {"Epoch": 0, "macro avg": {"f1": 0.7}, "loss": 0.04, "accuracy": 0.7},
        {"Epoch": 1, "macro avg": {"f1": 0.88}, "loss": 0.03, "accuracy": 0.8},
        {"Epoch": 1, "macro avg": {"f1": 0.7}, "loss": 0.02, "accuracy": 0.66}
    ]
    print(get_max(src, "accuracy", "macro avg:f1", with_keys="Epoch", merge=False))



# coding: utf-8
# 2021/3/15 @ tongshiwei
import re
from collections import OrderedDict
import torch


def save_params(filename, net: (torch.nn.Module, torch.nn.DataParallel), select=None):
    params = net.module.state_dict() if isinstance(net, torch.nn.DataParallel) else net.state_dict()

    if select is not None:
        pattern = re.compile(select)
        params = OrderedDict({name: value for name, value in params.items() if
                              pattern.match(name)})

    torch.save(params, filename)
    return filename



# coding: utf-8
# Copyright @tongshiwei

import torch

try:
    from .Module import *
except (SystemError, ModuleNotFoundError):
    from Module import *


class ModelName(object):
    def __init__(
            self, load_epoch=None, cfg=None, toolbox_init=False, **kwargs
    ):
        """
        模型初始化

        Parameters
        ----------
        load_epoch: int or None
            默认为 None，不为None时，将装载指定轮数的模型参数作为初始化模型参数
        cfg: Configuration, str or None
            默认为 None，不为None时，将使用cfg指定的参数配置
            或路径指定的参数配置作为模型参数
        toolbox_init: bool
            默认为 False，是否初始化工具包
        kwargs
            参数配置可选参数
        """
        # 1 配置参数初始化
        # todo 到Configuration处定义相关参数
        cfg = self.config(cfg, **kwargs)
        self.mod = self.get_module(cfg)

        # 2 todo 定义网络结构
        # 2.1 重新生成
        self.mod.logger.info("generating symbol")
        net = self.mod.sym_gen(
            **cfg.hyper_params
        )
        # 2.2 装载已有模型, export 出来的文件
        # net = mod.load(begin_epoch)
        # net = GluonModule.load_net(filename)

        self.net = net
        self.initialized = False

        if load_epoch is not None:
            self.model_init(load_epoch, allow_reinit=False)

        self.bp_loss_f = None
        self.loss_function = None
        self.toolbox = None
        self.trainer = None

        if toolbox_init:
            self.toolbox_init(**self.mod.cfg.toolbox_params)

    @staticmethod
    def config(cfg=None, **kwargs):
        """
        配置初始化

        Parameters
        ----------
        cfg: Configuration, str or None
            默认为 None，不为None时，将使用cfg指定的参数配置
            或路径指定的参数配置作为模型参数
        kwargs
            参数配置可选参数
        """
        cfg = Configuration(
            **kwargs
        ) if cfg is None else cfg
        if not isinstance(cfg, Configuration):
            cfg = Configuration.load_cfg(cfg, **kwargs)
        cfg.dump(override=True)
        return cfg

    @staticmethod
    def get_module(cfg):
        """
        根据配置，生成模型模块

        Parameters
        ----------
        cfg: Configuration
            模型配置参数
        Returns
        -------
        mod: Module
            模型模块
        """
        mod = Module(cfg)
        mod.logger.info(str(mod))
        filename = mod.cfg.cfg_path
        mod.logger.info("parameters saved to %s" % filename)
        return mod

    def set_loss(self, bp_loss_f=None, loss_function=None):
        bp_loss_f = get_bp_loss(self.mod.cfg.ctx, **self.mod.cfg.loss_params) if bp_loss_f is None else bp_loss_f

        assert bp_loss_f is not None and len(bp_loss_f) == 1

        loss_function = {

        } if loss_function is None else loss_function
        loss_function.update(bp_loss_f)

        self.bp_loss_f = bp_loss_f
        self.loss_function = loss_function

    def toolbox_init(
            self,
            evaluation_formatter_parameters=None,
            validation_logger_mode="w",
            silent=False,
    ):

        from longling import path_append
        from longling.lib.clock import Clock
        from longling.lib.utilog import config_logging
        from longling.ML.toolkit import EpochEvalFMT as Formatter
        from longling.ML.toolkit import MovingLoss, ConsoleProgressMonitor as ProgressMonitor

        self.toolbox = {
            "monitor": dict(),
            "timer": None,
            "formatter": dict(),
        }

        mod = self.mod
        cfg = self.mod.cfg

        # 4.1 todo 定义损失函数
        # bp_loss_f 定义了用来进行 back propagation 的损失函数，
        # 有且只能有一个，命名中不能为 *_\d+ 型

        assert self.loss_function is not None

        loss_monitor = MovingLoss(self.loss_function)

        # 4.1 todo 初始化一些训练过程中的交互信息
        timer = Clock()

        progress_monitor = ProgressMonitor(
            indexes={
                "Loss": [name for name in self.loss_function]
            },
            values={
                "Loss": loss_monitor.losses
            },
            end_epoch=cfg.end_epoch - 1,
            silent=silent
        )

        validation_logger = config_logging(
            filename=path_append(cfg.model_dir, "result.log"),
            logger="%s-validation" % cfg.model_name,
            mode=validation_logger_mode,
            log_format="%(message)s",
        )

        # set evaluation formatter
        evaluation_formatter_parameters = {} \
            if evaluation_formatter_parameters is None \
            else evaluation_formatter_parameters

        evaluation_formatter = Formatter(
            logger=validation_logger,
            dump_file=mod.cfg.validation_result_file,
            **evaluation_formatter_parameters
        )

        self.toolbox["monitor"]["loss"] = loss_monitor
        self.toolbox["monitor"]["progress"] = progress_monitor
        self.toolbox["timer"] = timer
        self.toolbox["formatter"]["evaluation"] = evaluation_formatter

    def model_init(
            self,
            load_epoch=None, force_init=False, cfg=None,
            allow_reinit=True, trainer=None,
            **kwargs
    ):
        mod = self.mod
        net = self.net
        cfg = mod.cfg if cfg is None else cfg
        begin_epoch = cfg.begin_epoch

        if self.initialized and not force_init:
            mod.logger.warning("model has been initialized, skip model_init")

        load_epoch = load_epoch if load_epoch is not None else begin_epoch

        # 5 todo 初始化模型
        model_file = kwargs.get(
            "init_model_file", mod.epoch_params_filename(load_epoch)
        )
        try:
            net = mod.load(load_epoch)
            mod.logger.info(
                "load params from existing model file "
                "%s" % model_file
            )
        except FileExistsError:
            if allow_reinit:
                mod.logger.info("model doesn't exist, initializing")
                self.Module.net_initialize(net, cfg.ctx)
            else:
                mod.logger.info(
                    "model doesn't exist, target file: %s" % model_file
                )

        self.initialized = True

        # # optional
        # # todo: whether to use static symbol to accelerate
        # # note: do not invoke this method for dynamic structure like rnn
        # # suggestion: annotate this until your process worked
        # net.hybridize()

        self.trainer = Module.get_trainer(
            net, optimizer=cfg.optimizer,
            optimizer_params=cfg.optimizer_params,
            lr_params=cfg.lr_params,
            select=cfg.train_select
        ) if trainer is None else trainer

    def train_net(self, train_data, eval_data, trainer=None):
        mod = self.mod
        cfg = self.mod.cfg
        net = self.net

        bp_loss_f = self.bp_loss_f
        loss_function = self.loss_function
        toolbox = self.toolbox

        batch_size = mod.cfg.batch_size
        begin_epoch = mod.cfg.begin_epoch
        end_epoch = mod.cfg.end_epoch

        assert all([bp_loss_f, loss_function]), \
            "make sure these variable have been initialized, " \
            "check init method and " \
            "make sure package_init method has been called"

        # 6 todo 训练
        trainer = self.trainer if trainer is None else trainer
        mod.logger.info("start training")
        mod.fit(
            net=net, begin_epoch=begin_epoch, end_epoch=end_epoch,
            batch_size=batch_size,
            train_data=train_data,
            trainer=trainer, bp_loss_f=bp_loss_f,
            loss_function=loss_function,
            eval_data=eval_data,
            ctx=None,
            toolbox=toolbox,
            prefix=mod.prefix,
            save_epoch=cfg.save_epoch,
        )

        # optional
        # # need execute the net.hybridize() before export,
        # # and execute forward at least one time
        # 需要在这之前调用 hybridize 方法,并至少forward一次
        # # net.export(mod.prefix)

    def step_fit(self, batch_data, trainer=None, loss_monitor=None):
        mod = self.mod
        net = self.net

        bp_loss_f = self.bp_loss_f
        loss_function = self.loss_function

        if loss_monitor is None:
            toolbox = self.toolbox
            monitor = toolbox.get("monitor")
            loss_monitor = monitor.get("loss") if monitor else None

        trainer = self.trainer if trainer is None else trainer
        mod.fit_f(
            net=net, batch_data=batch_data,
            trainer=trainer,
            bp_loss_f=bp_loss_f, loss_function=loss_function,
            loss_monitor=loss_monitor,
        )

    def __call__(self, *args, **kwargs):
        raise NotImplementedError

    def call(self, x, ctx=None):
        # call forward for single data

        # optional
        # pre process x

        # convert the data to ndarray
        x = torch.tensor([x])

        # forward
        outputs = self.net(x).asnumpy().tolist()[0]

        return outputs

    def batch_call(self, x):
        # call forward for batch data
        # notice: do not use too big batch size

        # optional
        # pre process x

        # convert the data to ndarray
        x = torch.tensor(x)

        # forward
        outputs = self.net(x).asnumpy().tolist()

        return outputs

    def transform(self, data):
        return transform(data, self.mod.cfg)

    def etl(self, data_src, cfg=None):
        mod = self.mod
        cfg = mod.cfg if cfg is None else cfg

        # 4.2 todo 定义数据加载
        mod.logger.info("loading data")
        data = etl(cfg.var2val(data_src), params=cfg)

        return data

    def _train(self, train, valid):
        train_data = self.etl(train)
        valid_data = self.etl(valid)
        self.train_net(train_data, valid_data)

    @staticmethod
    def train(*args, cfg=None, **kwargs):
        module = ModelName(cfg=cfg, **kwargs)
        module.set_loss()
        # module.viz()

        module.toolbox_init()
        module.model_init(**kwargs)

        module._train(*args)

        return module

    @staticmethod
    def test(test_filename, test_epoch, dump_file=None, **kwargs):
        from longling.ML.toolkit.formatter import EpochEvalFMT as Formatter
        formatter = Formatter(dump_file=dump_file)
        module = ModelName.load(test_epoch, **kwargs)

        test_data = module.etl(test_filename)
        eval_result = module.mod.eval(module.net, test_data)
        formatter(
            tips="test",
            eval_name_value=eval_result
        )
        return eval_result

    @staticmethod
    def inc_train(init_model_file, *args, validation_logger_mode="w", **kwargs):
        # 增量学习，从某一轮参数继续训练
        module = ModelName(**kwargs)
        module.toolbox_init(validation_logger_mode=validation_logger_mode)
        module.model_init(init_model_file=init_model_file)

        module._train(*args)

    @staticmethod
    def dump_configuration(**kwargs):
        ModelName.get_module(**kwargs)

    @staticmethod
    def load(load_epoch=None, **kwargs):
        module = ModelName(**kwargs)
        load_epoch = module.mod.cfg.end_epoch if load_epoch is None \
            else load_epoch
        module.model_init(load_epoch, **kwargs)
        return module

    @staticmethod
    def get_parser():
        cfg_parser = ConfigurationParser(
            Configuration,
            commands=[
                ModelName.config,
                ModelName.train, ModelName.test,
                ModelName.inc_train,
            ]
        )
        return cfg_parser

    @staticmethod
    def run(parse_args=None):
        cfg_parser = ModelName.get_parser()
        cfg_kwargs = cfg_parser(parse_args)

        if "subcommand" not in cfg_kwargs:
            cfg_parser.print_help()
            return
        subcommand = cfg_kwargs["subcommand"]
        del cfg_kwargs["subcommand"]

        eval("%s.%s" % (ModelName.__name__, subcommand))(**cfg_kwargs)


if __name__ == '__main__':
    ModelName.run()



# coding: utf-8
# 2020/1/7 @ tongshiwei


class ServiceModule(object):  # pragma: no cover
    def __init__(self, cfg=None, **kwargs):
        self.cfg = self.config(cfg, **kwargs)
        self.mod = self.get_module(self.cfg)

        self.loss_function = None

    @staticmethod
    def get_configuration_cls():
        raise NotImplementedError

    @staticmethod
    def get_module_cls():
        raise NotImplementedError

    @classmethod
    def config(cls, cfg=None, **kwargs):
        """
        配置初始化

        Parameters
        ----------
        cfg: Configuration, str or None
            默认为 None，不为None时，将使用cfg指定的参数配置
            或路径指定的参数配置作为模型参数
        kwargs
            参数配置可选参数
        """
        configuration_cls = cls.get_configuration_cls()

        cfg = configuration_cls(
            **kwargs
        ) if cfg is None else cfg
        if not isinstance(cfg, configuration_cls):
            cfg = configuration_cls.load_cfg(cfg, **kwargs)
        return cfg

    @classmethod
    def get_module(cls, cfg):
        """
        根据配置，生成模型模块

        Parameters
        ----------
        cfg: Configuration
            模型配置参数
        Returns
        -------
        mod: Module
            模型模块
        """
        module_cls = cls.get_module_cls()

        mod = module_cls(cfg)
        mod.logger.info(str(mod))
        filename = mod.cfg.cfg_path
        mod.logger.info("parameters saved to %s" % filename)
        return mod


class CliServiceModule(ServiceModule):  # pragma: no cover
    @staticmethod
    def get_configuration_cls():
        raise NotImplementedError

    @staticmethod
    def get_module_cls():
        raise NotImplementedError

    @staticmethod
    def get_configuration_parser_cls():
        raise NotImplementedError

    @classmethod
    def cli_commands(cls):
        raise NotImplementedError

    @classmethod
    def get_parser(cls):
        configuration_parser_cls = cls.get_configuration_parser_cls()
        configuration_cls = cls.get_configuration_cls()

        cfg_parser = configuration_parser_cls(
            configuration_cls,
            commands=cls.cli_commands()
        )
        return cfg_parser

    @classmethod
    def run(cls, parse_args=None):
        cfg_parser = cls.get_parser()
        cfg_kwargs = cfg_parser(parse_args)

        if "subcommand" not in cfg_kwargs:
            cfg_parser.print_help()
            return
        subcommand = cfg_kwargs["subcommand"]
        del cfg_kwargs["subcommand"]

        eval("cls.%s" % subcommand)(**cfg_kwargs)


def service_wrapper(
        meta_model_cls: type(CliServiceModule),
        configuration_cls=None,
        configuration_parser_cls=None,
        module_cls=None):  # pragma: no cover
    class MetaModel(meta_model_cls):
        @staticmethod
        def get_configuration_cls():
            if configuration_cls is not None:
                return configuration_cls
            else:
                return meta_model_cls.get_configuration_cls()

        @staticmethod
        def get_module_cls():
            if module_cls is not None:
                return module_cls
            else:
                return meta_model_cls.get_module_cls()

        @staticmethod
        def get_configuration_parser_cls():
            if configuration_parser_cls is not None:
                return configuration_parser_cls
            else:
                return meta_model_cls.get_configuration_parser_cls()

    return MetaModel



# coding: utf-8
# create by tongshiwei on 2020-11-16
import os
from longling import ConfigurationParser

__all__ = ["Configuration", "ConfigurationParser", "directory_check"]

import datetime

from longling import path_append
from longling.lib import parser
from longling.lib.parser import var2exp
from longling.lib.utilog import config_logging, LogLevel
from ..const import CFG_JSON, RESULT_JSON


class Configuration(parser.Configuration):
    # 目录配置
    model_name = "MLModel"

    root = "./"
    dataset = ""
    timestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
    workspace = ""

    root_data_dir = path_append("$root", "data", "$dataset") if dataset else path_append("$root", "data")
    data_dir = path_append("$root_data_dir", "data")
    root_model_dir = path_append("$root_data_dir", "model", "$model_name")
    model_dir = path_append("$root_model_dir", "$workspace") if workspace else root_model_dir
    cfg_path = path_append("$model_dir", "configuration.json")

    root = str(root)
    root_data_dir = str(root_data_dir)
    root_model_dir = str(root_model_dir)

    # 说明
    caption = ""

    def __init__(self, params_path=None, params_kwargs=None, **kwargs):
        """
        Configuration File, including categories:

        * directory setting
        * optimizer setting
        * training parameters
        * equipment
        * parameters saving setting
        * user parameters

        Parameters
        ----------
        params_path: str
            The path to configuration file which is in json format
        kwargs:
            Parameters to be reset.

        Notes
        -----
        Developers modifying this code should simultaneously modify the relevant codes in glue and tore

        Examples
        --------
        >>> cfg = Configuration(epoch=5)
        >>> cfg["epoch"]
        5
        >>> cfg   # doctest: +ELLIPSIS
        logger: <Logger MLModel (INFO)>
        model_name: MLModel
        root: ./
        ...
        """
        super(Configuration, self).__init__()

        params = self.class_var
        if params_path:
            params.update(self.load_cfg(cfg_path=params_path, **(params_kwargs if params_kwargs else {})))
        params.update(**kwargs)

        self.validation_result_file = None
        self.cfg_path = None

        self._update(**params)

    def _update(self, **kwargs):
        params = kwargs

        params["logger"] = params.pop(
            "logger",
            config_logging(
                logger=params.get("model_name", self.model_name),
                console_log_level="info"
            )
        )

        for key in params:
            if key.endswith("_params") and key + "_update" in params:
                params[key].update(params[key + "_update"])

        # path_override_check
        path_check_list = ["dataset", "root_data_dir", "workspace", "root_model_dir", "model_dir"]
        _overridden = {}
        for path_check in path_check_list:
            if kwargs.get(path_check) is None or kwargs[path_check] == getattr(self, "%s" % path_check):
                _overridden[path_check] = False
            else:
                _overridden[path_check] = True

        for param, value in params.items():
            setattr(self, "%s" % param, value)

        def is_overridden(varname):
            return _overridden["%s" % varname]

        # set dataset
        if is_overridden("dataset") and not is_overridden("root_data_dir"):
            kwargs["root_data_dir"] = path_append("$root", "data", "$dataset")
        # set workspace
        if (is_overridden("workspace") or is_overridden("root_model_dir")) and not is_overridden("model_dir"):
            kwargs["model_dir"] = path_append("$root_model_dir", "$workspace")

        # rebuild relevant directory or file path according to the kwargs
        _dirs = [
            "workspace", "root_data_dir", "data_dir", "root_model_dir",
            "model_dir"
        ]
        for _dir in _dirs:
            exp = var2exp(
                kwargs.get(_dir, getattr(self, _dir)),
                env_wrap=lambda x: "self.%s" % x
            )
            setattr(self, _dir, eval(exp))

        self.validation_result_file = path_append(
            self.model_dir, RESULT_JSON, to_str=True
        )
        self.cfg_path = path_append(
            self.model_dir, CFG_JSON, to_str=True
        )

    def dump(self, cfg_path=None, override=True, file_format=None):
        cfg_path = self.cfg_path if cfg_path is None else cfg_path
        super(Configuration, self).dump(cfg_path, override, file_format=file_format)

    def var2val(self, var):
        return eval(var2exp(
            var,
            env_wrap=lambda x: "self.%s" % x
        ))


def directory_check(class_obj):
    print("root", os.path.abspath(class_obj.root))


if __name__ == '__main__':
    # Advise: firstly checkout whether the directory is correctly (step 1) and
    # then generate the paramters configuation file
    # to check the details (step 2)
    stage = 1
    # step 1
    directory_check(Configuration)

    if stage > 1:
        # 命令行参数配置
        _kwargs = ConfigurationParser.get_cli_cfg(Configuration)

        cfg = Configuration(
            **_kwargs
        )
        print(_kwargs)
        print(cfg)

        if stage > 2:
            # # step 2
            cfg.dump(override=True)
            try:
                logger = cfg.logger
                cfg.load_cfg(cfg.cfg_path)
                cfg.logger = logger
                cfg.logger.info('format check done')
            except Exception as e:
                print("parameters format error, may contain illegal data type")
                raise e



# coding: utf-8
# 2021/8/5 @ tongshiwei

from .utils import getF, copy_net, array_equal
from .sequence import get_begin_state, mask_sequence_variable_length, format_sequence
from .embedding import set_embedding_weight



# coding: utf-8
# 2019/9/21 @ tongshiwei

__all__ = ["main_cli", "docs_cli", "dockerfile_cli", "gitlab_ci_cli"]

import datetime
import os
import pathlib
from collections import OrderedDict
from copy import deepcopy

from longling.Architecture.config import STAGE_CANDIDATES
from longling.Architecture.install_proj import project_types
from longling.Architecture.utils import legal_input, binary_legal_input, default_legal_input


def main_cli(skip_top, project, **kwargs):
    if skip_top:
        default_project_name = pathlib.PurePath(os.getcwd()).name
        project = legal_input("Project Name (default is %s) < " % default_project_name,
                              __default_value="%s" % default_project_name) if not project else project
    else:
        project = legal_input("Project Name < ", is_legal=lambda x: True if x else False) if not project else project
    params = dict(
        project=legal_input("Project Name < ", is_legal=lambda x: True if x else False) if not project else project,
        project_type=legal_input("Project Type (%s) < " % "/".join(project_types), __legal_input=project_types),
    )
    return params


def docs_cli(project=None, title=None, author=None, copyright=None, default_style="sphinx",
             **kwargs):
    params = dict(
        project=legal_input("Project Name < ", is_legal=lambda x: True if x else False) if not project else project,
        docs_style=default_legal_input("Docs Style",
                                       __legal_input={"mxnet", "sphinx"},
                                       __default_value="%s" % default_style),
        docs_root="docs/" if binary_legal_input("Make 'docs/' directory?", _default="y") else "./",
    )
    if params["docs_style"] != "sphinx":
        params.update(dict(
            title=legal_input("Docs Title (default is %s) < " % params["project"],
                              __default_value=params["project"]) if not title else title,
            author=legal_input("Author < ", is_legal=lambda x: True if x else False) if not author else author,

        ))

        default_copyright = "%s, %s" % (datetime.date.today().year, params["author"])
        params.update(dict(
            copyright=legal_input("Copyright (default is %s) < " % default_copyright,
                                  __default_value=default_copyright) if not copyright else copyright,
        ))
    params.update(readthedocs=binary_legal_input("Install .readthedocs.yml?"))

    if binary_legal_input("Install Dockerfile for documents?"):
        params.update(dict(
            docker_params=dockerfile_cli("docs", docker_type="nginx"))
        )

    return params


def dockerfile_cli(project_type, docker_type=None, port=None):
    project_docker = {
        "python": {"cli", "flask"},
        "web": {"nginx", "vue"},
        "docs": {"nginx"},
    }

    docker_params = {}

    docker_type = legal_input(
        "Choose a service type (%s) < " % "/".join(project_docker[project_type]),
        __legal_input=project_docker[project_type],
    ) if docker_type is None else docker_type

    docker_params["docker_type"] = docker_type

    if docker_type in {"nginx", "vue"}:
        docker_params.update(
            image_name=default_legal_input(
                "Choose a image", __default_value="nginx"
            ),
            path_to_html=default_legal_input(
                "Specify the html directory",
                __default_value="_build/html" if project_type == "docs" else "build/html"
            )
        )

    if project_type == "python":
        docker_params.update(
            image_name=default_legal_input(
                "Choose a image", __default_value="python:3.6"
            )
        )
        if docker_type == "cli":
            docker_params.update(
                path_to_main=default_legal_input(
                    "Specify the main entry (e.g., the path to main.py)",
                    is_legal=lambda x: True if x else False
                )
            )
        elif docker_type == "flask":
            docker_params.update(
                path_to_main=default_legal_input(
                    "Specify the main entry (e.g., main_package.main_py:main_func)",
                    is_legal=lambda x: True if x else False
                ),
                port=default_legal_input(
                    "Specify the port that docker will listen",
                    is_legal=lambda x: True if x else False
                ) if port is None else port
            )

        else:  # pragma: no cover
            raise TypeError("%s: cannot handle docker type %s, only supports %s" % (
                project_type, docker_type, project_docker[project_type]))

    return docker_params


def gitlab_ci_cli(port=None, stages_candidates: (OrderedDict, dict) = None, docs=False, **kwargs):
    def get_stage_params(stage_image_name=None, stage_image_port=None, on_stop=False, only_master=None, manual=None):
        _params = {}
        if stage_image_name is None:
            _params["image_name"] = default_legal_input("Choose a image", is_legal=lambda x: True if x else False)
        else:
            _params["image_name"] = default_legal_input("Choose a image", __default_value=stage_image_name)

        if stage_image_port is not None:
            _params["image_port"] = legal_input("Stage Image Port (default is %s) < " % stage_image_port,
                                                __default_value=stage_image_port)
        if on_stop:
            _params["on_stop"] = binary_legal_input("Add Corresponding Stop Stage?")
        if only_master is not None:
            _params["only_master"] = binary_legal_input("Only triggered in master branch?", _default=only_master)
        if manual is not None:
            _params["manual"] = binary_legal_input("Triggered manually?", _default=manual)
        return _params

    _stages_candidates = deepcopy(STAGE_CANDIDATES)

    if stages_candidates is not None:
        _stages_candidates.update(stages_candidates)

    if docs is False:
        _stages_candidates.pop("docs")

    deployment_image = None

    params = {}
    for stage, ques_params in _stages_candidates.items():
        if binary_legal_input("Need [%s] Stage?" % stage, _default=ques_params.get("need", "y")):
            if stage in {"review", "production"}:
                if "image_port" not in ques_params and port is not None:
                    ques_params.update({"stage_image_port": port})
                if "stage_image_name" not in ques_params and deployment_image is not None:
                    ques_params.update({"stage_image_name": deployment_image})
            params[stage] = get_stage_params(**ques_params)
            if stage in {"review", "production"} and "image_name" in params[stage]:
                deployment_image = params[stage]["image_name"]
            elif stage == "build" and "image_name" in params[stage]:
                _stages_candidates["test"]["stage_image_name"] = params[stage]["image_name"]

    return params



# coding: utf-8
# create by tongshiwei on 2019-9-4


__all__ = ["Module"]


class Module(object):
    """
    Examples
    --------
    >>> class Params:
    ...     def __init__(self):
    ...         self.a = 1
    ...         self.b = 2
    >>> params = Params()
    >>> print(Module(params))
    Params
    a: 1
    b: 2
    """

    def __init__(self, parameters):
        self.cfg = parameters

    def __str__(self):
        """
        显示模块参数
        Display the necessary params of this Module

        Returns
        -------

        """
        string = ["Params"]
        for k, v in vars(self.cfg).items():
            string.append("%s: %s" % (k, v))
        return "\n".join(string)



# coding: utf-8
# 2021/8/5 @ tongshiwei

from mxnet import symbol, ndarray
from mxnet.gluon.parameter import tensor_types

from longling.lib.candylib import as_list

__all__ = ["format_sequence", "mask_sequence_variable_length", "get_begin_state"]


def format_sequence(length, inputs, layout, merge, in_layout=None):
    """
    `Original Code <https://github.com/apache/incubator-mxnet/blob/master/python/mxnet/gluon/rnn/rnn_cell.py#L52>`_

    Parameters
    ----------
    length
    inputs
    layout
    merge
    in_layout

    Returns
    -------

    Examples
    --------
    >>> import mxnet.ndarray as nd
    >>> seq = [[[0] * 4, [2] * 4, [4] * 4], [[1] * 4, [3] * 4, [5] * 4]]
    >>> seq1, axis, _, batch_size = format_sequence(3, nd.array(seq), "NTC", False)
    >>> seq1   # doctest: +NORMALIZE_WHITESPACE
    [
    [[0. 0. 0. 0.]
     [1. 1. 1. 1.]]
    <NDArray 2x4 @cpu(0)>,
    [[2. 2. 2. 2.]
     [3. 3. 3. 3.]]
    <NDArray 2x4 @cpu(0)>,
    [[4. 4. 4. 4.]
     [5. 5. 5. 5.]]
    <NDArray 2x4 @cpu(0)>]
    >>> axis
    1
    >>> batch_size
    2
    >>> seq2, _, _, _ = format_sequence(3, nd.array(seq), "NTC", True)
    >>> seq2   # doctest: +NORMALIZE_WHITESPACE
    <BLANKLINE>
    [[[0. 0. 0. 0.]
      [2. 2. 2. 2.]
      [4. 4. 4. 4.]]
    <BLANKLINE>
     [[1. 1. 1. 1.]
      [3. 3. 3. 3.]
      [5. 5. 5. 5.]]]
    <NDArray 2x3x4 @cpu(0)>
    >>> import mxnet.symbol as sym
    >>> seq3, _, _, _ = format_sequence(3, sym.Variable("s", shape=(2, 3, 4)), "NTC", False)
    >>> seq3
    [<Symbol split0>, <Symbol split0>, <Symbol split0>]
    >>> seq4 = [nd.array([[0] * 4, [1] * 4]), nd.array([[2] * 4, [3] * 4]), nd.array([[4] * 4, [5] * 4])]
    >>> seq5, _, _, _ = format_sequence(3, seq4, "NTC", True)
    >>> seq5   # doctest: +NORMALIZE_WHITESPACE
    <BLANKLINE>
    [[[0. 0. 0. 0.]
      [2. 2. 2. 2.]
      [4. 4. 4. 4.]]
    <BLANKLINE>
     [[1. 1. 1. 1.]
      [3. 3. 3. 3.]
      [5. 5. 5. 5.]]]
    <NDArray 2x3x4 @cpu(0)>
    >>> seq6 = [sym.Variable("1", shape=(2, 4)), sym.Variable("2", shape=(2, 4)), sym.Variable("3", shape=(2, 4))]
    >>> seq7, _, _, _ = format_sequence(3, seq6, "NTC", True)
    >>> seq7
    <Symbol stack0>
    """
    assert inputs is not None, \
        "unroll(inputs=None) has been deprecated. " \
        "Please create input variables outside unroll."

    axis = layout.find('T')
    batch_axis = layout.find('N')
    batch_size = 0
    in_axis = in_layout.find('T') if in_layout is not None else axis
    if isinstance(inputs, symbol.Symbol):
        F = symbol
        if merge is False:
            assert len(inputs.list_outputs()) == 1, \
                "unroll doesn't allow grouped symbol as input. " \
                "Please convert " \
                "to list with list(inputs) first or " \
                "let unroll handle splitting."
            inputs = list(symbol.split(
                inputs, axis=in_axis, num_outputs=length, squeeze_axis=1
            ))
    elif isinstance(inputs, ndarray.NDArray):
        F = ndarray
        batch_size = inputs.shape[batch_axis]
        if merge is False:
            assert length is None or length == inputs.shape[in_axis]
            inputs = as_list(ndarray.split(inputs, axis=in_axis,
                                           num_outputs=inputs.shape[in_axis],
                                           squeeze_axis=1))
    else:
        assert length is None or len(inputs) == length
        if isinstance(inputs[0], symbol.Symbol):
            F = symbol
        else:
            F = ndarray
            batch_size = inputs[0].shape[batch_axis]
        if merge is True:
            inputs = F.stack(*inputs, axis=axis)
            in_axis = axis

    if isinstance(inputs, tensor_types) and axis != in_axis:  # pragma: no cover
        # todo: find the test case
        inputs = F.swapaxes(inputs, dim1=axis, dim2=in_axis)

    return inputs, axis, F, batch_size


def mask_sequence_variable_length(F, data, length, valid_length, time_axis,
                                  merge):
    """
    `Original Code <https://github.com/apache/incubator-mxnet/blob/master/python/mxnet/gluon/rnn/rnn_cell.py#L82>`_

    Parameters
    ----------
    F
    data
    length
    valid_length
    time_axis
    merge: bool

    Returns
    -------
    masked_sequence: list, ...
        if merge is False, return list of step vector

    Examples
    --------
    >>> import mxnet.ndarray as nd
    >>> import mxnet as mx
    >>> mask_sequence_variable_length(
    ...     nd, mx.nd.ones((2, 4, 3)), 4, nd.array([2, 4]), 1, False
    ... )    # doctest: +NORMALIZE_WHITESPACE
    [
    [[1. 1. 1.]
     [1. 1. 1.]]
    <NDArray 2x3 @cpu(0)>,
    [[1. 1. 1.]
     [1. 1. 1.]]
    <NDArray 2x3 @cpu(0)>,
    [[0. 0. 0.]
     [1. 1. 1.]]
    <NDArray 2x3 @cpu(0)>,
    [[0. 0. 0.]
     [1. 1. 1.]]
    <NDArray 2x3 @cpu(0)>]
    >>> mask_sequence_variable_length(
    ...     nd, mx.nd.ones((2, 4, 3)), 4, nd.array([2, 4]), 1, True
    ... )    # doctest: +NORMALIZE_WHITESPACE
    <BLANKLINE>
    [[[1. 1. 1.]
      [1. 1. 1.]
      [0. 0. 0.]
      [0. 0. 0.]]
    <BLANKLINE>
     [[1. 1. 1.]
      [1. 1. 1.]
      [1. 1. 1.]
      [1. 1. 1.]]]
    <NDArray 2x4x3 @cpu(0)>
    >>> mask_sequence_variable_length(
    ...     nd, [mx.nd.ones((2, 3)), mx.nd.ones((2, 3)), mx.nd.ones((2, 3)), mx.nd.ones((2, 3))],
    ...     4, nd.array([2, 4]), 1, True
    ... )
    <BLANKLINE>
    [[[1. 1. 1.]
      [1. 1. 1.]
      [0. 0. 0.]
      [0. 0. 0.]]
    <BLANKLINE>
     [[1. 1. 1.]
      [1. 1. 1.]
      [1. 1. 1.]
      [1. 1. 1.]]]
    <NDArray 2x4x3 @cpu(0)>
    """
    assert valid_length is not None
    if not isinstance(data, tensor_types):
        data = F.stack(*data, axis=time_axis)
    outputs = F.SequenceMask(
        data, sequence_length=valid_length,
        use_sequence_length=True,
        axis=time_axis
    )
    if not merge:
        outputs = as_list(
            F.split(
                outputs, num_outputs=length, axis=time_axis, squeeze_axis=True
            )
        )
    return outputs


def get_begin_state(cell, F, begin_state, inputs, batch_size):
    """
    `Original Code <https://github.com/apache/incubator-mxnet/blob/master/python/mxnet/gluon/rnn/rnn_cell.py#L45>`_

    Parameters
    ----------
    cell
    F
    begin_state
    inputs
    batch_size

    Returns
    -------
    >>> import mxnet.ndarray as nd
    >>> from mxnet import gluon
    >>> lstm_cell = gluon.rnn.LSTMCell(3)
    >>> get_begin_state(lstm_cell, nd, None, nd.ones((2, 4, 5)), 2)   # doctest: +NORMALIZE_WHITESPACE
    [
    [[0. 0. 0.]
     [0. 0. 0.]]
    <NDArray 2x3 @cpu(0)>,
    [[0. 0. 0.]
     [0. 0. 0.]]
    <NDArray 2x3 @cpu(0)>]
    >>> import mxnet.symbol as sym
    >>> get_begin_state(lstm_cell, sym, None,
    ...     sym.Variable("inputs", shape=(2, 4, 5)), 2)   # doctest: +NORMALIZE_WHITESPACE
    [<Symbol lstm0_begin_state_2>, <Symbol lstm0_begin_state_3>]
    """
    if begin_state is None:
        if F is ndarray:
            ctx = inputs.context if isinstance(inputs, tensor_types) else inputs[0].context
            with ctx:
                begin_state = cell.begin_state(func=F.zeros, batch_size=batch_size)
        else:
            begin_state = cell.begin_state(func=F.zeros, batch_size=batch_size)
    return begin_state



# coding: utf-8
# 2021/8/3 @ tongshiwei

"""
Here are some frequently used regex expression for select in collect_params()
"""
# 包含所有参数
ALL_PARAMS = None

# 除外所有embedding层
BLOCK_EMBEDDING = "^(?!.*embedding)"



# coding: utf-8
# 2021/3/14 @ tongshiwei

import functools


def eval_wrapper(_eval_f):
    @functools.wraps(_eval_f)
    def eval_f(_net, *args, **kwargs):
        _net.eval()
        ret = _eval_f(_net, *args, **kwargs)
        _net.train()
        return ret

    return eval_f



# coding: utf-8
# 2019/12/20 @ tongshiwei

import warnings

warnings.warn("unstable, not use")
"""
mem_info.total: 显卡总的显存大小
mem_info.used: 使用显存大小
mem_info.free: 剩余显存大小

上述三个量单位都是字节bytes
"""
try:
    import pynvml

    pynvml.nvmlInit()
except Exception as e:  # pragma: no cover
    import logging

    logging.error("import pynvml failed")
    logging.error(e)
    logging.warning("do not use this module before the problem is fixed")


def _get_device_info(device_id):
    handle = pynvml.nvmlDeviceGetHandleByIndex(device_id)
    mem_info = pynvml.nvmlDeviceGetMemoryInfo(handle)

    return mem_info


def _get_device_util(device_id):
    mem_info = _get_device_info(device_id)

    return mem_info.used / mem_info.total


def get_device_util(device_id=None):
    if device_id is not None:
        return _get_device_util(device_id)
    else:
        return [
            _get_device_util(i) for i in range(pynvml.nvmlDeviceGetCount())
        ]


def _get_device_free(device_id):
    mem_info = _get_device_info(device_id)

    return mem_info.free / mem_info.total


def get_device_free(device_id: int = None):
    if device_id is not None:
        return _get_device_free(device_id)
    else:
        return [
            _get_device_free(i) for i in range(pynvml.nvmlDeviceGetCount())
        ]


def get_free_device_ids(threshold=0.5, device_ids=None):
    device_ids = set(device_ids) if device_ids else set(range(pynvml.nvmlDeviceGetCount()))
    return [_id for _id, _f in get_device_free() if _f >= threshold and _id in device_ids]


def get_device_num():
    return pynvml.nvmlDeviceGetCount()



# coding: utf-8
from __future__ import absolute_import

import logging
import time
from contextlib import contextmanager

from longling.lib.utilog import config_logging

_logger = config_logging(logger="clock", console_log_level=logging.INFO)

__all__ = ["Clock", "print_time", "Timer"]


@contextmanager
def print_time(tips: str = "", logger=_logger):
    """
    统计并打印脚本运行时间， 秒为单位

    Parameters
    ----------
    tips: str
    logger: logging.Logger or logging

    Examples
    --------
    >>> with print_time("tips"):
    ...     a = 1 + 1  # The code you want to test

    """
    start_time = time.time()
    logger.info('Starting to %s', tips)
    yield
    logger.info('Finished to {} in {:.6f} seconds'.format(
        tips,
        time.time() - start_time)
    )


class Clock(object):
    r"""
    计时器。
    包含两种时间：wall_time 和 process_time

    * wall_time: 包括等待时间在内的程序运行时间
    * process_time: 不包括等待时间在内的程序运行时间

    Parameters
    ----------
    store_dict: dict or None
        with closure 中存储运行时间
    logger: logging.logger
        日志
    tips: str
        提示前缀

    Examples
    --------
    .. code-block :: python

        with Clock():
            a = 1 + 1
        clock = Clock()
        clock.start()
        # some code
        clock.end(wall=True) # default to return the wall_time, to get process_time, set wall=False

    """

    def __init__(self, store_dict: (dict, None) = None, logger: (logging.Logger, None) = _logger, tips=''):
        assert store_dict is None or type(store_dict) is dict
        self.process_st = 0
        self.process_et = 0
        self.wall_st = 0
        self.wall_et = 0
        self.store_dict = store_dict
        self.logger = logger
        self.tips = tips

    def start(self):
        """开始计时"""
        self.process_st = time.process_time()
        self.wall_st = time.time()
        return self.process_st

    def time(self):
        return time.time() - self.wall_st

    def end(self, wall=True):
        """计时结束，返回间隔时间"""
        self.process_et = time.process_time()
        self.wall_et = time.time()
        if wall:
            return self.wall_time
        else:
            return self.process_time

    @property
    def wall_time(self):
        """获取程序运行时间（包括等待时间）"""
        return self.wall_et - self.wall_st

    @property
    def process_time(self):
        """获取程序运行时间（不包括等待时间）"""
        return self.process_et - self.process_st

    def __enter__(self):
        if self.tips:
            if self.logger is not None:
                self.logger.info(self.tips)
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.logger is not None:
            self.logger.info(
                '%ss' % self.end() if not self.tips else '%s %ss' % (
                    self.tips, self.end()
                )
            )
        if self.store_dict is not None:
            self.store_dict['wall_time'] = self.wall_time
            self.store_dict['process_time'] = self.process_time


Timer = Clock



# coding: utf-8
# 2019/12/10 @ tongshiwei

"""
Spider parser, targeting at html, parse the component
"""

from bs4 import BeautifulSoup


def as_bs4(lxml_text: (str, BeautifulSoup)) -> BeautifulSoup:
    return lxml_text if isinstance(lxml_text, BeautifulSoup) else BeautifulSoup(lxml_text)


__all__ = ["get_all_url"]


def get_all_url(lxml_text) -> list:
    return [a.get('href') for a in as_bs4(lxml_text).find_all('a')]


def get_all_text(lxml_text) -> str:
    return as_bs4(lxml_text).get_text()



# coding: utf-8
# 2021/3/15 @ tongshiwei
import logging
from torch.optim.lr_scheduler import *


def get_lr_scheduler(scheduler: (str, ...) = "cosine", logger=logging, update_params=None, adaptive=True, *, optimizer,
                     **kwargs):
    """

    Parameters
    ----------
    scheduler
    logger
    update_params
    adaptive
    optimizer
    kwargs

    Returns
    -------

    Examples
    --------
    >>> import torch
    >>> scheduler = get_lr_scheduler(
    ...     "OneCycleLR",
    ...     max_lr=1,
    ...     total_steps=100,
    ...     optimizer=torch.optim.Adam(torch.nn.Linear(10, 30).parameters(), 0.1)
    ... )
    >>> scheduler.get_last_lr()[0]  # doctest: +ELLIPSIS
    0.040...
    >>> for _ in range(50):
    ...     scheduler.step()
    >>> scheduler.get_last_lr()[0]  # doctest: +ELLIPSIS
    0.793...
    >>> for _ in range(50):
    ...     scheduler.step()
    >>> scheduler.get_last_lr()[0]  # doctest: +ELLIPSIS
    0.0005...
    """
    if isinstance(scheduler, str):
        scheduler = eval(scheduler)(optimizer=optimizer, **kwargs)

    logger.info(scheduler)

    return scheduler


def plot_schedule(scheduler, iterations=1500):  # pragma: no cover
    import matplotlib.pyplot as plt
    lrs = []
    for _ in range(iterations):
        scheduler.step()
        lrs.append(scheduler.get_last_lr()[0])
    plt.scatter(list(range(iterations)), lrs)
    plt.xlabel("Iteration")
    plt.ylabel("Learning Rate")
    plt.show()


if __name__ == '__main__':
    import torch

    logging.getLogger().setLevel(logging.INFO)
    _scheduler = get_lr_scheduler(
        "OneCycleLR", max_lr=1, total_steps=1500, optimizer=torch.optim.Adam(torch.nn.Linear(10, 30).parameters(), 0.1)
    )
    plot_schedule(_scheduler)



# coding: utf-8
# create by tongshiwei on 2020-11-13


__all__ = ["get_default_toolbox"]


def get_default_toolbox(
        loss_function=None,
        evaluation_formatter_parameters=None,
        progress_monitor_parameters=None,
        validation_logger_mode="w",
        silent=False,
        configuration=None
):  # pragma: no cover
    """
    New in version 1.3.16

    todo: consider whether to keep it

    Notice
    ------
    The developer who modify this document should simultaneously modify the related function in glue
    """

    from longling import path_append
    from longling.lib.clock import Clock
    from longling.lib.utilog import config_logging
    from longling.ML.toolkit import EpochEvalFMT as Formatter
    from longling.ML.toolkit import MovingLoss, ConsoleProgressMonitor as ProgressMonitor

    cfg = configuration

    toolbox = {
        "monitor": dict(),
        "timer": None,
        "formatter": dict(),
    }

    loss_monitor = MovingLoss(loss_function) if loss_function else None

    timer = Clock()

    progress_monitor = ProgressMonitor(
        indexes={
            "Loss": [name for name in loss_function]
        } if loss_function else {},
        values={
            "Loss": loss_monitor.losses
        } if loss_monitor else {},
        silent=silent,
        **progress_monitor_parameters if progress_monitor_parameters is not None else {}
    )

    validation_logger = config_logging(
        filename=path_append(cfg.model_dir, "result.log") if hasattr(cfg, "model_dir") else None,
        logger="%s-validation" % cfg.model_name if hasattr(cfg, "model_name") else "model",
        mode=validation_logger_mode,
        log_format="%(message)s",
    )

    # set evaluation formatter
    evaluation_formatter_parameters = {} \
        if evaluation_formatter_parameters is None \
        else evaluation_formatter_parameters

    evaluation_formatter = Formatter(
        logger=validation_logger,
        dump_file=getattr(cfg, "validation_result_file", False),
        **evaluation_formatter_parameters
    )

    toolbox["monitor"]["loss"] = loss_monitor
    toolbox["monitor"]["progress"] = progress_monitor
    toolbox["timer"] = timer
    toolbox["formatter"]["evaluation"] = evaluation_formatter

    return toolbox



# coding: utf-8
# 2020/1/2 @ tongshiwei

import logging
import pathlib
from tqdm import tqdm
from longling import as_out_io, as_io, PATH_IO_TYPE, PATH_TYPE, IO_TYPE
import csv
import json

logger = logging.getLogger("longling")

__all__ = ["csv2jsonl", "jsonl2csv", "loading", "load_jsonl", "load_csv", "load_file"]


def csv2jsonl(src: PATH_IO_TYPE, tar: PATH_IO_TYPE = None, delimiter=",", **kwargs):
    """
    将 csv 格式文件/io流 转换为 json 格式文件/io流

    transfer csv file or io stream into  json file or io stream

    Parameters
    ----------
    src: PATH_IO_TYPE
        数据源，可以是文件路径，也可以是一个IO流。
        the path to source file or io stream.
    tar: PATH_IO_TYPE
        输出目标，可以是文件路径，也可以是一个IO流。
        the path to target file or io stream.
    delimiter: str
        分隔符
        the delimiter used in csv. some usually used delimiters are ","  and " "
    kwargs: dict
        options passed to csv.DictWriter

    Examples
    --------
    Assume such component is written in demo.csv:

    .. code-block::

        column1,column2
        hello,world
        hello,you

    use following codes to reading the component

    .. code-block:: python

        csv2json("demo.csv", "demo.jsonl")

    and get

    .. code-block::

        {'column1': 'hello', 'column2': 'world'}
        {'column1': 'hello', 'column2': 'you'}
    """
    with as_out_io(tar) as wf:
        for line in tqdm(load_csv(src, delimiter=delimiter, **kwargs), "csv2json: %s --> %s" % (src, tar)):
            print(json.dumps(line, ensure_ascii=False), file=wf)


def jsonl2csv(src: PATH_IO_TYPE, tar: PATH_IO_TYPE = None, delimiter=",", **kwargs):
    """
    将 json 格式文件/io流 转换为 csv 格式文件/io流

    transfer json file or io stream into csv file or io stream

    Parameters
    ----------
    src: PATH_IO_TYPE
        数据源，可以是文件路径，也可以是一个IO流。
        the path to source file or io stream.
    tar: PATH_IO_TYPE
        输出目标，可以是文件路径，也可以是一个IO流。
        the path to target file or io stream.
    delimiter: str
        分隔符
        the delimiter used in csv. some usually used delimiters are ","  and " "
    kwargs: dict
        options passed to csv.DictWriter

    Examples
    --------
    Assume such component is written in demo.csv:

    .. code-block::

        {'column1': 'hello', 'column2': 'world'}
        {'column1': 'hello', 'column2': 'you'}

    use following codes to reading the component

    .. code-block:: python

        jsonl2csv("demo.csv", "demo.jsonl")

    and get

    .. code-block::

        column1,column2
        hello,world
        hello,you
    """
    with as_out_io(tar) as wf:
        csv_writer = None
        for line in tqdm(load_jsonl(src), "json2csv: %s --> %s" % (src, tar)):
            if csv_writer is None:
                csv_writer = csv.DictWriter(wf, line.keys(), delimiter=delimiter, **kwargs)
                csv_writer.writeheader()
            csv_writer.writerow(line)


def load_file(src: PATH_IO_TYPE):
    """
    Read raw text from source

    Examples
    --------
    Assume such component is written in demo.txt:

    .. code-block::

        hello
        world

    use following codes to reading the component

    .. code-block:: python

        for line in load_csv('demo.txt'):
            print(line, end="")

    and get

    .. code-block::

        hello
        world
    """
    with as_io(src) as f:
        for line in f:
            yield line


def load_csv(src: PATH_IO_TYPE, delimiter=",", **kwargs):
    """
    read the dict from csv

    Examples
    --------

    Assume such component is written in demo.csv:

    .. code-block::

        a,b,c
        1,2,3
        2,4,6


    .. code-block:: python

        for line in load_csv('demo.csv'):
            print(line)


    .. code-block::

        {"a": 1, "b": 2, "c": 3}
        {"a": 2, "b": 4, "c": 6}
    """
    with as_io(src) as f:
        field_names = [i for i in csv.reader([f.readline()], delimiter=delimiter, **kwargs)][0]

        for line in csv.DictReader(f, field_names, delimiter=delimiter, **kwargs):
            yield line


def load_jsonl(src: PATH_IO_TYPE):
    """
    缓冲式按行读取jsonl文件

    Examples
    --------

    Assume such component is written in demo.jsonl:

    .. code-block::

        {"a": 1}
        {"a": 2}


    .. code-block:: python

        for line in load_jsonl('demo.jsonl'):
            print(line)


    .. code-block::

        {"a": 1}
        {"a": 2}
    """
    with as_io(src) as f:
        for line in f:
            yield json.loads(line)


def loading(src: (PATH_IO_TYPE, ...), src_type=None):
    """
    缓冲式按行读取文件

    Support read from

    * jsonl (apply load_jsonl)
    * csv (apply load_csv).
    * file in other format will be treated as raw text (apply load_file).
    * function will be invoked and return
    * others will be directly returned
    """
    if isinstance(src, PATH_TYPE):
        suffix = pathlib.PurePath(src).suffix[1:]
        src_type = src_type if src_type is not None else suffix
        if src_type in {"csv", ".csv"}:
            return load_csv(src)
        elif src_type in {"json", "jsonl", ".json", ".jsonl"}:
            if src_type == "json":
                logger.info("detect source type as json, processed as jsonl format")
            return load_jsonl(src)
        else:
            return load_file(src)
    elif isinstance(src, IO_TYPE):
        return load_file(src)
    elif callable(src):
        return src()
    return src



# coding: utf-8
# create by tongshiwei on 2019/7/2

import logging

import os

from urllib.request import urlretrieve
from .utils import decompress, reporthook4urlretrieve as _reporthook4urlretrieve

logger = logging.getLogger("spider")

__all__ = ["download_file"]


def download_file(url, save_path=None, override=True, decomp=True, reporthook=None):
    """
    cli alias: ``download``, download data from specified url

    Parameters
    ----------
    url
    save_path
    override
    decomp
    reporthook

    Returns
    -------

    """
    save_path = url.split('/')[-1] if not save_path else save_path
    if os.path.exists(save_path):
        if override is True:
            os.remove(save_path)
            logger.warning(save_path + ' will be overridden.')
        else:
            raise FileExistsError("%s existed, downloading abandoned" % save_path)

    logger.info(url + ' is saved as %s', save_path)
    if reporthook is None:
        urlretrieve(url, save_path, reporthook=_reporthook4urlretrieve)
        print()
    else:
        urlretrieve(url, save_path, reporthook=reporthook)
    if decomp:
        decompress(save_path)



# coding: utf-8
# 2020/5/10 @ tongshiwei

import warnings
from longling import loading, print_time

__all__ = ["to_board"]

from .utils import get_by_key


def to_board(src, board_dir, global_step_field, *scalar_fields):
    with warnings.catch_warnings(record=True):
        from tensorboardX import SummaryWriter

    with SummaryWriter(board_dir) as sw, print_time(
            "to_board: %s -> %s\n step field: %s, fields: %s" % (src, board_dir, global_step_field, scalar_fields)
    ):
        for line in loading(src):
            for scalar_field in scalar_fields:
                sw.add_scalar(tag=scalar_field, scalar_value=get_by_key(line, scalar_field),
                              global_step=int(get_by_key(line, global_step_field)))



# coding: utf-8
# 2020/4/13 @ tongshiwei

# incubating

import logging
import json
from longling.lib.stream import as_out_io, PATH_IO_TYPE
from collections import OrderedDict
from longling.lib.formatter import table_format, series_format
from longling.lib.candylib import as_list
import warnings

__all__ = ["eval_format", "EvalFMT", "EpochEvalFMT", "EpisodeEvalFMT", "result_format"]


def _to_dict(name_value: (dict, tuple)) -> dict:
    """Make sure the name_value a dict object"""
    return dict(name_value) if isinstance(name_value, tuple) else name_value


class EvalFMT(object):
    """
    评价指标格式化类。可以按一定格式快速格式化评价指标。

    Parameters
    ----------
    logger:
        默认为 root logger
    dump_file:
        不为空时，将结果写入dump_file
    col: int
        每行放置的指标数量
    kwargs:
        拓展兼容性参数

    Examples
    --------
    >>> import numpy as np
    >>> from longling.ML.metrics import classification_report
    >>> y_true = np.array([0, 0, 1, 1, 2, 1])
    >>> y_pred = np.array([2, 1, 0, 1, 1, 0])
    >>> y_score = np.array([
    ...     [0.15, 0.4, 0.45],
    ...     [0.1, 0.9, 0.0],
    ...     [0.33333, 0.333333, 0.333333],
    ...     [0.15, 0.4, 0.45],
    ...     [0.1, 0.9, 0.0],
    ...     [0.33333, 0.333333, 0.333333]
    ... ])
    >>> print(EvalFMT.format(
    ...     iteration=30,
    ...     eval_name_value=classification_report(y_true, y_pred, y_score)
    ... ))    # doctest: +NORMALIZE_WHITESPACE
    Iteration [30]
               precision    recall        f1  support
    0           0.000000  0.000000  0.000000        2
    1           0.333333  0.333333  0.333333        3
    2           0.000000  0.000000  0.000000        1
    macro_avg   0.111111  0.111111  0.111111        6
    accuracy: 0.166667	macro_auc: 0.194444
    """

    def __init__(self, logger=logging.getLogger(), dump_file: (PATH_IO_TYPE, None) = False,
                 col: (int, None) = None, **kwargs):
        """

        Parameters
        ----------

        """
        self.logger = logger
        if dump_file is not False:
            # clean file
            with as_out_io(dump_file):
                pass
        self.log_f = dump_file
        self.col = col

    @classmethod
    def _loss_format(cls, name, value):
        return "%s: %s" % (name, value)

    def loss_format(self, loss_name_value):
        msg = []
        for name, value in loss_name_value.items():
            msg.append("Loss - " + self._loss_format(name, value))
        return " ".join(msg), loss_name_value

    @classmethod
    def format(cls, tips: str = None,
               iteration: int = None, train_time: float = None, loss_name_value: dict = None,
               eval_name_value: dict = None,
               extra_info: (dict, tuple) = None, keep: (set, str) = "msg",
               logger=logging.getLogger(), dump_file: (PATH_IO_TYPE, None) = False,
               col: (int, None) = None,
               *args, **kwargs):
        return cls(logger=logger, dump_file=dump_file, col=col)(
            tips=tips, iteration=iteration, train_time=train_time, loss_name_value=loss_name_value,
            eval_name_value=eval_name_value, extra_info=extra_info,
            dump=dump_file is not False, keep=keep, *args, **kwargs
        )

    @property
    def iteration_name(self):
        return "Iteration"

    @property
    def iteration_fmt(self):
        return self.iteration_name + " [{:d}]"

    def __call__(self, tips: str = None,
                 iteration: int = None, train_time: float = None, loss_name_value: dict = None,
                 eval_name_value: dict = None,
                 extra_info: (dict, tuple) = None,
                 dump: bool = True, keep: (set, str) = "data", *args, **kwargs):
        msg = []
        data = {}

        if tips is not None:
            msg.append("%s" % tips)

        if iteration is not None:
            msg.append(self.iteration_fmt.format(iteration))
            data[self.iteration_name] = iteration

        if train_time is not None:
            msg.append("Train Time-%.3fs" % train_time)
            data['train_time'] = train_time

        if loss_name_value is not None:
            loss_name_value = _to_dict(loss_name_value)
            assert isinstance(
                loss_name_value, dict
            ), "loss_name_value should be None, dict or tuple, " \
               "now is %s" % type(loss_name_value)
            _msg, _data = self.loss_format(loss_name_value)

            msg.append(
                _msg
            )
            data.update(
                _data
            )

        if extra_info is not None:
            extra_info = _to_dict(extra_info)
            assert isinstance(
                extra_info, dict
            ), "extra_info should be None, dict or tuple, " \
               "now is %s" % type(extra_info)
            msg.append(str(extra_info))
            data.update(extra_info)

        msg = ["\t".join([m for m in msg if m])]

        if eval_name_value is not None:
            eval_name_value = _to_dict(eval_name_value)
            assert isinstance(
                eval_name_value, dict
            ), "eval_name_value should be None, dict or tuple, " \
               "now is %s" % type(eval_name_value)
            msg.append(
                result_format(eval_name_value, col=self.col)
            )
            data.update(
                eval_name_value
            )

        msg = "\n".join([m for m in msg if m])

        if dump:
            logger = kwargs.get('logger', self.logger)
            logger.info("\n" + msg)
            log_f = kwargs.get('log_f', self.log_f)
            if log_f is not False:
                try:
                    with as_out_io(log_f, "a") as wf:
                        print(json.dumps(data, ensure_ascii=False), file=wf)
                except Exception as e:  # pragma: no cover
                    warnings.warn("Result dumping to file aborted: %s" % str(e))

        if keep is None:
            return msg
        elif isinstance(keep, str):
            keep = set(as_list(keep))

        if "msg" in keep and "data" in keep:
            return msg, data
        elif "msg" in keep:
            return msg
        elif "data" in keep:
            return data


def result_format(data: dict, col=None):
    """

    Parameters
    ----------
    data
    col

    Returns
    -------

    Examples
    --------
    >>> print(result_format({"a": 1, "b": 2}))    # doctest: +NORMALIZE_WHITESPACE
    a: 1	b: 2
    >>> print(result_format({"a": 1, "b": {"1": 0.1, "2": 0.3}, "c": {"1": 0.4, "2": 0.0}}))
         1    2
    b  0.1  0.3
    c  0.4  0.0
    a: 1
    """
    table = OrderedDict()
    series = OrderedDict()
    for key, value in data.items():
        if isinstance(value, dict):
            table[key] = value
        else:
            series[key] = value

    _ret = []
    if table:
        _ret.append(table_format(table))
    if series:
        _ret.append(series_format(series, col=col))

    return "\n".join(_ret)


eval_format = EvalFMT.format


class EpochEvalFMT(EvalFMT):
    """
    Examples
    --------
    >>> import numpy as np
    >>> from longling.ML.metrics import classification_report
    >>> y_true = np.array([0, 0, 1, 1, 2, 1])
    >>> y_pred = np.array([2, 1, 0, 1, 1, 0])
    >>> y_score = np.array([
    ...     [0.15, 0.4, 0.45],
    ...     [0.1, 0.9, 0.0],
    ...     [0.33333, 0.333333, 0.333333],
    ...     [0.15, 0.4, 0.45],
    ...     [0.1, 0.9, 0.0],
    ...     [0.33333, 0.333333, 0.333333]
    ... ])
    >>> print(EpochEvalFMT.format(
    ...     iteration=30,
    ...     eval_name_value=classification_report(y_true, y_pred, y_score)
    ... ))    # doctest: +NORMALIZE_WHITESPACE
    Epoch [30]
               precision    recall        f1  support
    0           0.000000  0.000000  0.000000        2
    1           0.333333  0.333333  0.333333        3
    2           0.000000  0.000000  0.000000        1
    macro_avg   0.111111  0.111111  0.111111        6
    accuracy: 0.166667	macro_auc: 0.194444
    """

    @property
    def iteration_name(self):
        return "Epoch"


class EpisodeEvalFMT(EvalFMT):
    """
    Examples
    --------
    >>> import numpy as np
    >>> from longling.ML.metrics import classification_report
    >>> y_true = np.array([0, 0, 1, 1, 2, 1])
    >>> y_pred = np.array([2, 1, 0, 1, 1, 0])
    >>> y_score = np.array([
    ...     [0.15, 0.4, 0.45],
    ...     [0.1, 0.9, 0.0],
    ...     [0.33333, 0.333333, 0.333333],
    ...     [0.15, 0.4, 0.45],
    ...     [0.1, 0.9, 0.0],
    ...     [0.33333, 0.333333, 0.333333]
    ... ])
    >>> print(EpisodeEvalFMT.format(
    ...     iteration=30,
    ...     eval_name_value=classification_report(y_true, y_pred, y_score)
    ... ))    # doctest: +NORMALIZE_WHITESPACE
    Episode [30]
               precision    recall        f1  support
    0           0.000000  0.000000  0.000000        2
    1           0.333333  0.333333  0.333333        3
    2           0.000000  0.000000  0.000000        1
    macro_avg   0.111111  0.111111  0.111111        6
    accuracy: 0.166667	macro_auc: 0.194444
    """

    @property
    def iteration_name(self):
        return "Episode"



# coding: utf-8
# 2019/9/20 @ tongshiwei

r"""
To replace the variable str which is in the pattern r"(\$[A-z_]*)\b" （i.e., `$xxx`)
"""

__all__ = ["variable_replace", "default_variable_replace"]

import re

VARIABLE_REGEX = r"(\$[A-z_]*)"
VARIABLE = re.compile(VARIABLE_REGEX)


def variable_replace(string: str, key_lower: bool = True, quotation: str = "", **variables):
    r"""
    Examples
    --------
    >>> string = "hello $who"
    >>> variable_replace(string, who="world")
    'hello world'
    >>> string = "hello $WHO"
    >>> variable_replace(string, key_lower=False, WHO="world")
    'hello world'
    >>> string = "hello $WHO"
    >>> variable_replace(string, who="longling")
    'hello longling'
    >>> string = "hello $Wh_o"
    >>> variable_replace(string, wh_o="longling")
    'hello longling'
    """

    def replace(match):
        variable = string[match.start() + 1:match.end()]
        variable = variable.lower() if key_lower else variable
        variable = variables[variable]
        return quotation + variable + quotation

    return VARIABLE.sub(replace, string)


def default_variable_replace(string: str, default_value: (str, None, dict) = None, key_lower: bool = True,
                             quotation: str = "",
                             **variables) -> str:
    """
    Examples
    --------
    >>> string = "hello $who, I am $author"
    >>> default_variable_replace(string, default_value={"author": "groot"}, who="world")
    'hello world, I am groot'
    >>> string = "hello $who, I am $author"
    >>> default_variable_replace(string, default_value={"author": "groot"})
    'hello , I am groot'
    >>> string = "hello $who, I am $author"
    >>> default_variable_replace(string, default_value='', who="world")
    'hello world, I am '
    >>> string = "hello $who, I am $author"
    >>> default_variable_replace(string, default_value=None, who="world")
    'hello world, I am $author'
    """
    if default_value is None or isinstance(default_value, str):
        return _default_variable_replace(string, default_value, key_lower, quotation, **variables)
    elif isinstance(default_value, dict):
        return _dict_variable_replace(string, default_value, key_lower, quotation, **variables)
    else:
        raise TypeError("cannot handle the type %s for default_value" % type(default_value))


def _dict_variable_replace(string: str, default_value: dict, key_lower: bool = True, quotation: str = "",
                           **variables) -> str:
    def replace(match):
        variable = string[match.start() + 1:match.end()]
        variable = variable.lower() if key_lower else variable
        if variable not in variables:
            return default_value.get(variable, "")
        return quotation + str(variables[variable]) + quotation

    return VARIABLE.sub(replace, string)


def _default_variable_replace(string: str, default_value: (str, None) = "", key_lower: bool = True, quotation: str = "",
                              **variables) -> str:
    def replace(match):
        variable = string[match.start() + 1:match.end()]
        variable = variable.lower() if key_lower else variable
        if variable not in variables:
            if default_value is not None:
                return default_value
            else:
                variable = string[match.start():match.end()]
                variable = variable.lower() if key_lower else variable
                return variable
        return quotation + str(variables[variable]) + quotation

    return VARIABLE.sub(replace, string)



# coding: utf-8
# 2021/6/28 @ tongshiwei
from ..toolkit import result_format
from collections import OrderedDict

__all__ = ["POrderedDict"]


class POrderedDict(OrderedDict):
    """
    Examples
    --------
    >>> _d = {"a": 1, "b": {0: 2, 1: 3}, "c": {0: 1, 1: 5}}
    >>> _d
    {'a': 1, 'b': {0: 2, 1: 3}, 'c': {0: 1, 1: 5}}
    >>> POrderedDict(_d)
       0  1
    b  2  3
    c  1  5
    a: 1
    """

    def __repr__(self):
        return result_format(self)



# coding: utf-8
# create by tongshiwei on 2019/4/12
import functools
import logging

import mxnet as mx
from longling.ML import DL
from longling.ML.MxnetHelper.toolkit import get_trainer, load_net, save_params

__all__ = ["Module"]


class Module(DL.Module):
    def fit_f(self, *args, **kwargs):
        raise NotImplementedError

    def sym_gen(self, *args, **kwargs):
        raise NotImplementedError

    def load(self, net, epoch, ctx=mx.cpu(), allow_missing=False,
             ignore_extra=False):
        """"
        Load the existing net parameters

        Parameters
        ----------
        net: HybridBlock
            The network which has been initialized
            or loaded from the existed model
        epoch: str or int
            The epoch which specify the model
        ctx: Context or list of Context
                Defaults to ``mx.cpu()``.
        allow_missing: bool
        ignore_extra: bool

        Returns
        -------
        The initialized net
        """
        # 根据起始轮次装载已有的网络参数
        filename = self.epoch_params_filepath(epoch)
        return load_net(
            filename, net, ctx, allow_missing=allow_missing,
            ignore_extra=ignore_extra
        )

    def epoch_params_filepath(self, epoch):
        raise NotImplementedError

    def net_initialize(self, *args, **kwargs):
        """初始化网络参数"""
        raise NotImplementedError

    @staticmethod
    @functools.wraps(get_trainer)
    def get_trainer(net, optimizer='sgd', optimizer_params=None, lr_params=None, select=None, logger=logging,
                    *args, **kwargs):
        return get_trainer(
            net=net, optimizer=optimizer, optimizer_params=optimizer_params,
            lr_params=lr_params, select=select, logger=logger, *args, **kwargs
        )

    @staticmethod
    @functools.wraps(save_params)
    def save_params(filename, net, select):
        save_params(filename=filename, net=net, select=select)

    @staticmethod
    def eval(*args, **kwargs):
        raise NotImplementedError



# coding: utf-8
# 2020/3/10 @ tongshiwei

import functools
import time
import random
import json
from longling.lib.stream import as_io
from longling.lib.path import path_append, abs_current_dir
from longling.lib.candylib import as_list
from urllib.request import ProxyHandler, build_opener
from longling.spider.conf import logger

META_DATA = path_append(abs_current_dir(__file__), '../meta_data/')

HTTP_HEADER = {
    'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_1) '
                  'AppleWebKit/537.36 (KHTML, like Gecko) Chrome/78.0.3904.108 Safari/537.36',
    'Connection': 'keep-alive',
    'Sec-Fetch-Site': 'same-origin',
    'Sec-Fetch-Mode': 'cors',
}

with as_io(path_append(META_DATA, 'user_agents.txt')) as f:
    USER_AGENTS = [line.strip() for line in f.readlines()]

with as_io(path_append(META_DATA, 'IPpool.json')) as f:
    PROXIES = ["%s:%s" % (ip_port["ip"], ip_port["port"]) for ip_port in json.load(f)]


def get_proxies():
    return PROXIES


def get_http_header():
    # HTTP_HEADER.update(get_user_agent())
    return get_user_agent()


def get_opener():
    proxy_handler = ProxyHandler({
        'http': random.choice(get_proxies()),
    })
    opener = build_opener(proxy_handler)
    return opener


def get_user_agent():
    return {'User-Agent': random.choice(USER_AGENTS)}


def retry(max_retry=5, retry_interval=1, retry_errors=None, failed_exception=ConnectionError(), logger=logger):
    retry_errors = ConnectionResetError if retry_errors is None else retry_errors
    retry_errors = tuple(as_list(retry_errors))

    def retry_wrapper(f):
        @functools.wraps(f)
        def new_f(*args, **kwargs):
            for _ in range(max_retry):
                try:
                    return f(*args, **kwargs)
                except retry_errors as e:
                    time.sleep(retry_interval)
                    logger.debug(e)
            raise failed_exception

        return new_f

    return retry_wrapper



# coding: utf-8
# 2020/3/24 @ tongshiwei

import datetime

__all__ = ["get_current_timestamp"]


def get_current_timestamp() -> str:  # pragma: no cover
    """
    Examples
    ---------
    .. code-block :: python

        > get_current_timestamp()
        '20200327172235'

    """
    return datetime.datetime.now().strftime("%Y%m%d%H%M%S")



# coding: utf-8
# create by tongshiwei on 2019-9-4


from __future__ import absolute_import
from __future__ import print_function

import functools
import os

import longling
from longling.ML.DL import ecli
from longling.lib.parser import path_append

__all__ = ["cli"]

TORE_DIR = path_append(
    os.path.dirname(longling.__file__),
    "ML.PytorchHelper.tore.ModelName".replace(".", os.sep),
    to_str=True
)
cli = functools.partial(ecli, source_dir=TORE_DIR)

if __name__ == '__main__':
    cli()



# coding: utf-8
# create by tongshiwei on 2017/11/17


from . import candylib
from . import clock
from . import stream
from . import utilog
from . import regex
from . import progress

from .progress import ProgressMonitor, IterableMIcing



# coding: utf-8
# 2019/10/12 @ tongshiwei

from longling.lib.formatter import pandas_format, dict_format
from .select import get_max, get_min

__all__ = ["select_max", "arg_select_max"]


def _select(src, *keys, with_keys: (str, None) = None, with_all=False, func=None, **kwargs):
    res = func(src, *keys, with_keys=with_keys, with_all=with_all)
    if with_keys or with_all:
        print(pandas_format(res, **kwargs))
    else:
        print(dict_format(res, **kwargs))


def _select_max(src, *keys, with_keys: (str, None) = None, with_all=False, **kwargs):
    _select(src, *keys, with_keys=with_keys, with_all=with_all, func=get_max, **kwargs)


def _select_min(src, *keys, with_keys: (str, None) = None, with_all=False, **kwargs):
    _select(src, *keys, with_keys=with_keys, with_all=with_all, func=get_min, **kwargs)


def arg_select_max(*keys, src, with_keys=None, with_all=False, **kwargs):
    """
    cli alias: ``amax``
    """
    _select_max(src, *keys, with_keys=with_keys, with_all=with_all, **kwargs)


def select_max(src, *keys, with_keys=None, with_all=False, **kwargs):
    """
    cli alias: ``max``
    """
    _select_max(src, *keys, with_keys=with_keys, with_all=with_all, **kwargs)


def arg_select_min(*keys, src, with_keys=None, with_all=False, **kwargs):
    """
    cli alias: ``amin``
    """
    _select_min(src, *keys, with_keys=with_keys, with_all=with_all, **kwargs)


def select_min(src, *keys, with_keys=None, with_all=False, **kwargs):
    """
    cli alias: ``min``
    """
    _select_min(src, *keys, with_keys=with_keys, with_all=with_all, **kwargs)



# coding: utf-8
# create by tongshiwei on 2019/4/12

import mxnet as mx
from mxnet import gluon
from tqdm import tqdm

__all__ = ["extract", "transform", "etl", "pseudo_data_iter"]


# todo: define extract-transform-load process and implement the pesudo data iterator for testing

def pseudo_data_iter(_cfg):
    def pseudo_data_generation():
        # 在这里定义测试用伪数据流
        import random
        random.seed(10)

        raw_data = [
            [random.random() for _ in range(5)]
            for _ in range(1000)
        ]

        return raw_data

    return load(transform(pseudo_data_generation(), _cfg), _cfg)


def extract(data_src):
    raise NotImplementedError


def transform(raw_data, params):
    # 定义数据转换接口
    # raw_data --> batch_data

    batch_size = params.batch_size

    transformed_data = raw_data

    return transformed_data


def load(transformed_data, params):
    batch_size = params.batch_size

    return gluon.data.DataLoader(
        gluon.data.ArrayDataset(
            mx.nd.array(transformed_data, dtype="float32")
        ),
        batch_size
    )


def etl(*args, params):
    raw_data = extract(*args)
    transformed_data = transform(raw_data, params)
    load(transformed_data, params)
    raise NotImplementedError


if __name__ == '__main__':
    from longling.lib.structure import AttrDict
    import os

    filename = "../../data/data.json"
    print(os.path.abspath(filename))

    for data in tqdm(extract(filename)):
        pass

    parameters = AttrDict({"batch_size": 128})
    for data in tqdm(etl(filename, params=parameters)):
        pass



# coding: utf-8
# create by tongshiwei on 2019-9-4

from .ecli import cli as ecli
from .module import Module
from .optimizer_cfg import get_optimizer_cfg
from .service import CliServiceModule, service_wrapper
from .light_module import train
from .utils import *



# coding: utf-8
# create by tongshiwei on 2019-9-1
__all__ = ["fit_f", "eval_f"]

import torch
from tqdm import tqdm

from longling.ML.PytorchHelper import tensor2list


def _fit_f(_net, _data, bp_loss_f, loss_function, loss_monitor):
    data, label = _data
    output, _ = _net(data)
    bp_loss = None
    for name, func in loss_function.items():
        # todo modify the input to func
        loss = func(output, label)
        if name in bp_loss_f:
            bp_loss = loss
        loss_value = torch.mean(loss).item()
        if loss_monitor:
            loss_monitor.update(name, loss_value)
    return bp_loss


def eval_f(_net, test_data):
    ground_truth = []
    prediction = []

    def evaluation_function(y_true, y_pred):
        return 0

    for (data, data_mask, label, pick_index, label_mask) in tqdm(test_data, "evaluating"):
        with torch.no_grad():
            output, _ = _net(data, data_mask)
            pred = tensor2list(output)
            label = tensor2list(label)
            ground_truth.extend(label)
            prediction.extend(pred)
    return {
        "evaluation_name": evaluation_function(ground_truth, prediction)
    }


# #####################################################################
# ###     The following codes usually do not need modification      ###
# #####################################################################

def fit_f(net, batch_data,
          trainer, bp_loss_f, loss_function, loss_monitor=None
          ):
    """
    Defined how each step of batch train goes

    Parameters
    ----------
    net: HybridBlock
        The network which has been initialized
        or loaded from the existed model
    batch_data: Iterable
        The batch data for train
    trainer:
        The trainer used to update the parameters of the net
    bp_loss_f: dict with only one value and one key
        The function to compute the loss for the procession
        of back propagation
    loss_function: dict of function
        Some other measurement in addition to bp_loss_f
    loss_monitor: LossMonitor
        Default to ``None``
    ctx: Context or list of Context
        Defaults to ``mx.cpu()``.

    Returns
    -------

    """
    bp_loss = _fit_f(
        net, batch_data, bp_loss_f, loss_function, loss_monitor
    )
    assert bp_loss is not None
    trainer.zero_grad()
    bp_loss.backward()
    trainer.step()



# coding: utf-8
# created by tongshiwei on 18-1-27

import traceback

import mxnet as mx

from longling.lib.stream import build_dir

__all__ = ["VizError", "plot_network"]


class VizError(Exception):  # pragma: no cover
    pass


def plot_network(nn_symbol, save_path="plot/network", shape=None,
                 node_attrs=None, view=False):  # pragma: no cover
    node_attrs = {} if node_attrs is None else node_attrs

    graph = mx.viz.plot_network(nn_symbol, shape=shape, node_attrs=node_attrs)

    assert save_path
    build_dir(save_path)
    try:
        graph.render(save_path, view=view)
    except Exception:
        raise VizError(traceback.format_exc())



# coding:utf-8
# created by tongshiwei on 2018/10/9


# from .indicesor import Indicesor
# from .splitter.file_splitter import *
# from .splitter.iterable_splitter import *
# from .splitter.splitter import Splitter

from .feature import ID2Feature
from .sampler import *
from .split import train_test, train_valid_test, kfold



# coding: utf-8
# create by tongshiwei on 2020-11-21

from collections import Iterable

import pandas as pd
from longling import as_list
from numpy.random import default_rng
from tqdm import tqdm

"""
rating matrix
(query, key, value): value could be omit

contrastive triplet
(query, pos, neg)
"""

__all__ = ["Sampler", "TripletPairSampler", "UserSpecificPairSampler", "ItemSpecificSampler"]


class Sampler(object):
    def __init__(self, random_state=10, pad_value=0):
        self.random_state = default_rng(random_state)
        self.pad_value = pad_value

    def padding(self, sampled: list, n, *args, **kwargs) -> list:
        pad_value = kwargs.get("pad_value", self.pad_value)
        return sampled + [pad_value] * (n - len(sampled))


class TripletPairSampler(Sampler):
    """
    Examples
    --------
    >>> # implicit feedback
    >>> import pandas as pd
    >>> triplet_df = pd.DataFrame({
    ...     "query": [0, 1, 2],
    ...     "pos": [[1], [3, 0, 2], [1]],
    ...     "neg": [[], [], []]
    ... })
    >>> sampler = TripletPairSampler(triplet_df, "query", set_index=True)
    >>> rating_matrix = pd.DataFrame({
    ...     "query": [0, 1, 1, 1, 2],
    ...     "key": [1, 3, 0, 2, 1]
    ... })
    >>> triplet_df = TripletPairSampler.rating2triplet(rating_matrix, query_field="query", key_field="key")
    >>> triplet_df   # doctest: +NORMALIZE_WHITESPACE
                 pos neg
    query
    0            [1]  []
    1      [3, 0, 2]  []
    2            [1]  []
    >>> sampler = TripletPairSampler(triplet_df, "query")
    >>> sampler(0)
    (0, [0])
    >>> sampler(0, 3)
    (0, [0, 0, 0])
    >>> sampler(0, 3, padding=False)
    (0, [])
    >>> sampler = TripletPairSampler(triplet_df, "query", query_range=3, key_range=4)
    >>> sampler(0)
    (0, [0])
    >>> sampler(0, 3)
    (0, [0, 0, 0])
    >>> sampler(0, 3, padding=False)
    (0, [])
    >>> sampler(0, 5, padding=False, implicit=True)
    (3, [2, 3, 0])
    >>> sampler(0, 5, padding=False, implicit=True, excluded_key=[3])
    (2, [0, 2])
    >>> sampler(0, 5, padding=True, implicit=True, excluded_key=[3])
    (2, [2, 0, 0, 0, 0])
    >>> sampler(0, 5, implicit=True, pad_value=-1)
    (3, [2, 3, 0, -1, -1])
    >>> sampler(0, 5, implicit=True, fast_implicit=True, pad_value=-1)
    (3, [0, 2, 3, -1, -1])
    >>> sampler(0, 5, implicit=True, fast_implicit=True, with_n_implicit=3, pad_value=-1)
    (3, [0, 2, 3, -1, -1, -1, -1, -1])
    >>> sampler(0, 5, implicit=True, fast_implicit=True, with_n_implicit=3, pad_value=-1, padding_implicit=True)
    (3, [0, 2, 3, -1, -1, -1, -1, -1])
    >>> rating_matrix = pd.DataFrame({
    ...     "query": [0, 1, 1, 1, 2],
    ...     "key": [1, 3, 0, 2, 1],
    ...     "score": [1, 0, 1, 1, 0]
    ... })
    >>> triplet_df = TripletPairSampler.rating2triplet(
    ...     rating_matrix,
    ...     "query", "key",
    ...     value_field="score"
    ... )
    >>> triplet_df   # doctest: +NORMALIZE_WHITESPACE
                pos  neg
    query
    0           [1]   []
    1        [0, 2]  [3]
    2            []  [1]
    >>> sampler = TripletPairSampler(triplet_df, "query", query_range=3, key_range=4)
    >>> sampler([0, 1, 2], 5, implicit=True, pad_value=-1)
    [(3, [2, 3, 0, -1, -1]), (1, [1, -1, -1, -1, -1]), (3, [3, 0, 2, -1, -1])]
    >>> sampler([0, 1, 2], 5, pad_value=-1)
    [(0, [-1, -1, -1, -1, -1]), (1, [3, -1, -1, -1, -1]), (1, [1, -1, -1, -1, -1])]
    >>> sampler([0, 1, 2], 5, neg=False, pad_value=-1)
    [(1, [1, -1, -1, -1, -1]), (2, [0, 2, -1, -1, -1]), (0, [-1, -1, -1, -1, -1])]
    >>> sampler(rating_matrix["query"], 2, neg=rating_matrix["score"],
    ...     excluded_key=rating_matrix["key"], pad_value=-1)
    [(0, [-1, -1]), (2, [2, 0]), (1, [3, -1]), (1, [3, -1]), (0, [-1, -1])]
    >>> sampler(rating_matrix["query"], 2, neg=rating_matrix["score"],
    ...     excluded_key=rating_matrix["key"], pad_value=-1, return_column=True)
    ((0, 2, 1, 1, 0), ([-1, -1], [0, 2], [3, -1], [3, -1], [-1, -1]))
    >>> sampler(rating_matrix["query"], 2, neg=rating_matrix["score"],
    ...     excluded_key=rating_matrix["key"], pad_value=-1, return_column=True, split_sample_to_column=True)
    ((0, 2, 1, 1, 0), [(-1, 0, 3, 3, -1), (-1, 2, -1, -1, -1)])
    >>> rating_matrix = pd.DataFrame({
    ...     "query": [0, 1, 1, 1, 2],
    ...     "key": [1, 3, 0, 2, 1],
    ...     "score": [0.8, 0.4, 0.7, 0.5, 0.1]
    ... })
    >>> TripletPairSampler.rating2triplet(
    ...     rating_matrix,
    ...     "query", "key",
    ...     value_field="score",
    ...     value_threshold=0.5
    ... )   # doctest: +NORMALIZE_WHITESPACE
              pos  neg
    query
    0         [1]   []
    1      [0, 2]  [3]
    2          []  [1]
    """

    def __init__(self, triplet_df: pd.DataFrame,
                 query_field, pos_field="pos", neg_field="neg", set_index=False,
                 query_range: (int, tuple, list) = None, key_range: (int, tuple, list) = None,
                 random_state=10):
        super(TripletPairSampler, self).__init__(random_state)
        self.df = triplet_df
        self.query_field = query_field
        self.pos_field = pos_field
        self.neg_field = neg_field

        def _get_range(value: (None, int, tuple, list)) -> tuple:
            if isinstance(value, int):
                return 0, value
            else:
                return value

        self.query_range = _get_range(query_range)
        self.key_range = _get_range(key_range)
        if set_index:
            self.df.set_index(self.query_field, inplace=True)

    def __call__(self, query: (int, str, list), n=1, excluded_key=None, neg=True, implicit=False, padding=True,
                 *args, verbose=True, return_column=False, split_sample_to_column=False,
                 padding_implicit=False, fast_implicit=False, with_n_implicit=None, **kwargs):
        """
        To get only explicit neg (pos) samples, set implicit as False, padding as False,
        padding_implicit as False and with_n_implicit as None.

        To get at maximum of n samples, where the insufficient ones are implicit sampled,
        set padding and padding_implicit as True, and with_n_implicit as None.

        To get explicit samples with at maximum of n implicit samples,
        set implicit as False, with_n_implicit as n.

        Parameters
        ----------
        query
        n
        excluded_key
        neg: bool
            negative sampling (set neg as True) or positive sampling (set neg as False)
        implicit: bool
            explicit sampling (set implicit as False) or implicit sampling (set implicit as True)
        padding
        args
        verbose
        return_column
        split_sample_to_column
        padding_implicit
        fast_implicit
        with_n_implicit
        kwargs

        Returns
        -------

        """
        if isinstance(query, (int, str)):
            sample = self.sample(
                query=query,
                n=n,
                excluded_key=excluded_key,
                neg=neg
            ) if not implicit else self.implicit_sample(
                query=query,
                n=n,
                excluded_key=excluded_key,
                fast_mode=fast_implicit
            )
            n_sample = len(sample)
            padding_n = n
            if with_n_implicit is not None:
                padding_n += with_n_implicit
                implicit_sample = self.implicit_sample(
                    query=query, n=with_n_implicit, excluded_key=excluded_key,
                    samples=sample,
                    fast_mode=fast_implicit)
                n_sample += len(implicit_sample)
                sample += implicit_sample

            if padding is True:
                if padding_implicit is True:
                    if n_sample < padding_n:
                        padding_sample = self.implicit_sample(query=query, n=padding_n - n_sample,
                                                              excluded_key=excluded_key, samples=sample,
                                                              fast_mode=fast_implicit)
                        n_padding_sample = len(padding_sample)
                        return n_sample + n_padding_sample, self.padding(sample + padding_sample, padding_n, **kwargs)
                else:
                    return n_sample, self.padding(sample, padding_n, **kwargs)

            return n_sample, sample
        else:
            neg = neg if isinstance(neg, Iterable) else [neg] * len(query)
            excluded_key = excluded_key if isinstance(excluded_key, Iterable) else [excluded_key] * len(query)
            n_and_sample = [
                self(
                    _query, n, _excluded_key, _neg, implicit, padding,
                    *args,
                    padding_implicit=padding_implicit,
                    fast_implicit=fast_implicit,
                    with_n_implicit=with_n_implicit,
                    **kwargs)
                for _query, _excluded_key, _neg in tqdm(
                    zip(query, excluded_key, neg), "sampling", disable=not verbose
                )
            ]
            if return_column:
                n_sample, sample = zip(*n_and_sample)
                if split_sample_to_column:
                    return n_sample, list(zip(*sample))
                return n_sample, sample
            return n_and_sample

    def sample(self, query: (int, str, list), n=1, excluded_key=None, neg=True, *args, **kwargs):
        candidates = self.df.loc[query][self.neg_field] if neg else self.df.loc[query][self.pos_field]

        if excluded_key is not None:
            candidates = list(set(candidates) - set(as_list(excluded_key)))

        sampled = self.random_state.choice(candidates, min(n, len(candidates)), replace=False).tolist()
        # rs = default_rng(10)
        # sampled = rs.choice(candidates, min(n, len(candidates)), replace=False).tolist()
        return sampled

    def implicit_sample(self, query: (int, str, list), n=1, excluded_key=None, fast_mode=False, samples=None,
                        *args, fast_max_try=100, **kwargs):
        exclude = set(self.df.loc[query][self.pos_field]) | set(self.df.loc[query][self.neg_field])
        if excluded_key is not None:
            exclude |= set(as_list(excluded_key))
        if samples is not None:
            exclude |= set(samples)
        if fast_mode is False:
            candidates = list(set(range(*self.key_range)) - exclude)
            sampled = self.random_state.choice(candidates, min(n, len(candidates)), replace=False).tolist()
            return sampled
        else:
            sampled = set()
            try_cnt = 0
            while len(sampled) < n and try_cnt < n + fast_max_try:
                _sample = self.random_state.integers(*self.key_range)
                if _sample not in exclude and _sample not in sampled:
                    sampled.add(_sample)
                try_cnt += 1
            return list(sampled)

    @staticmethod
    def rating2triplet(rating_matrix: pd.DataFrame, query_field, key_field, value_field=None,
                       value_threshold=None,
                       pos_field="pos", neg_field="neg",
                       *args, verbose=True, **kwargs):
        if isinstance(rating_matrix, pd.DataFrame):
            triplet = []
            for _, group in tqdm(rating_matrix.groupby(query_field), "rating2triplet", disable=not verbose):
                query_key = group[query_field].unique()[0]
                if value_field is None:
                    pos = group[key_field]
                    neg = pd.Series([], dtype=pos.dtype)
                elif value_threshold:
                    pos = group[group[value_field] >= value_threshold][key_field]
                    neg = group[group[value_field] < value_threshold][key_field]
                else:
                    pos = group[group[value_field] == 1][key_field]
                    neg = group[group[value_field] == 0][key_field]
                triplet.append([query_key, pos.tolist(), neg.tolist()])
        else:
            raise TypeError()

        df = pd.DataFrame(triplet, columns=[query_field, pos_field, neg_field])
        return df.set_index(query_field)


class UserSpecificPairSampler(TripletPairSampler):
    """
    Examples
    --------
    >>> import pandas as pd
    >>> user_num = 3
    >>> item_num = 4
    >>> rating_matrix = pd.DataFrame({
    ...     "user_id": [0, 1, 1, 1, 2],
    ...     "item_id": [1, 3, 0, 2, 1]
    ... })
    >>> triplet_df = UserSpecificPairSampler.rating2triplet(rating_matrix)
    >>> triplet_df   # doctest: +NORMALIZE_WHITESPACE
                   pos neg
    user_id
    0              [1]  []
    1        [3, 0, 2]  []
    2              [1]  []
    >>> sampler = UserSpecificPairSampler(triplet_df)
    >>> sampler(1)
    (0, [0])
    >>> sampler = UserSpecificPairSampler(triplet_df, item_id_range=item_num)
    >>> sampler(0, implicit=True)
    (1, [3])
    >>> sampler(0, 5, implicit=True)
    (3, [3, 2, 0, 0, 0])
    >>> sampler(0, 5, implicit=True, pad_value=-1)
    (3, [3, 2, 0, -1, -1])
    >>> sampler([0, 1, 2], 5, implicit=True, pad_value=-1)
    [(3, [2, 3, 0, -1, -1]), (1, [1, -1, -1, -1, -1]), (3, [2, 0, 3, -1, -1])]
    >>> rating_matrix = pd.DataFrame({
    ...     "user_id": [0, 1, 1, 1, 2],
    ...     "item_id": [1, 3, 0, 2, 1],
    ...     "score": [1, 0, 1, 1, 0]
    ... })
    >>> triplet_df = UserSpecificPairSampler.rating2triplet(rating_matrix=rating_matrix, value_field="score")
    >>> triplet_df   # doctest: +NORMALIZE_WHITESPACE
                pos  neg
    user_id
    0           [1]   []
    1        [0, 2]  [3]
    2            []  [1]
    >>> sampler = UserSpecificPairSampler(triplet_df)
    >>> sampler([0, 1, 2], 5, pad_value=-1)
    [(0, [-1, -1, -1, -1, -1]), (1, [3, -1, -1, -1, -1]), (1, [1, -1, -1, -1, -1])]
    >>> sampler([0, 1, 2], 5, neg=False, pad_value=-1)
    [(1, [1, -1, -1, -1, -1]), (2, [0, 2, -1, -1, -1]), (0, [-1, -1, -1, -1, -1])]
    >>> sampler(rating_matrix["user_id"], 2, neg=rating_matrix["score"],
    ...     excluded_key=rating_matrix["item_id"], pad_value=-1)
    [(0, [-1, -1]), (2, [2, 0]), (1, [3, -1]), (1, [3, -1]), (0, [-1, -1])]
    >>> sampler(rating_matrix["user_id"], 2, neg=rating_matrix["score"],
    ...     excluded_key=rating_matrix["item_id"], pad_value=-1, return_column=True)
    ((0, 2, 1, 1, 0), ([-1, -1], [0, 2], [3, -1], [3, -1], [-1, -1]))
    >>> sampler(rating_matrix["user_id"], 2, neg=rating_matrix["score"],
    ...     excluded_key=rating_matrix["item_id"], pad_value=-1, return_column=True, split_sample_to_column=True)
    ((0, 2, 1, 1, 0), [(-1, 2, 3, 3, -1), (-1, 0, -1, -1, -1)])
    """

    def __init__(self, triplet_df: pd.DataFrame,
                 query_field="user_id", pos_field="pos", neg_field="neg", set_index=False,
                 user_id_range=None, item_id_range=None,
                 random_state=10):
        super(UserSpecificPairSampler, self).__init__(
            triplet_df=triplet_df,
            query_field=query_field,
            pos_field=pos_field,
            neg_field=neg_field,
            set_index=set_index,
            query_range=user_id_range,
            key_range=item_id_range,
            random_state=random_state)

    @staticmethod
    def rating2triplet(rating_matrix: pd.DataFrame,
                       query_field="user_id", key_field="item_id", value_field=None,
                       value_threshold=None, pos_field="pos", neg_field="neg",
                       *args, verbose=True, **kwargs):
        return TripletPairSampler.rating2triplet(
            rating_matrix=rating_matrix,
            query_field=query_field,
            key_field=key_field,
            value_field=value_field,
            value_threshold=value_threshold,
            pos_field=pos_field,
            neg_field=neg_field,
            *args,
            verbose=verbose,
            **kwargs
        )


class ItemSpecificSampler(TripletPairSampler):
    """
    Examples
    --------
    >>> import pandas as pd
    >>> user_num = 3
    >>> item_num = 4
    >>> rating_matrix = pd.DataFrame({
    ...     "user_id": [0, 1, 1, 1, 2],
    ...     "item_id": [1, 3, 0, 2, 1]
    ... })
    >>> triplet_df = ItemSpecificSampler.rating2triplet(rating_matrix)
    >>> triplet_df   # doctest: +NORMALIZE_WHITESPACE
                pos neg
    item_id
    0           [1]  []
    1        [0, 2]  []
    2           [1]  []
    3           [1]  []
    >>> triplet_df.index
    Int64Index([0, 1, 2, 3], dtype='int64', name='item_id')
    >>> sampler = ItemSpecificSampler(triplet_df)
    >>> sampler(1)
    (0, [0])
    >>> sampler = ItemSpecificSampler(triplet_df, user_id_range=user_num)
    >>> sampler(0, implicit=True)
    (1, [2])
    >>> sampler(0, 5, implicit=True)
    (2, [2, 0, 0, 0, 0])
    >>> sampler(0, 5, implicit=True, pad_value=-1)
    (2, [0, 2, -1, -1, -1])
    >>> sampler([0, 1, 2], 5, implicit=True, pad_value=-1)
    [(2, [0, 2, -1, -1, -1]), (1, [1, -1, -1, -1, -1]), (2, [0, 2, -1, -1, -1])]
    >>> rating_matrix = pd.DataFrame({
    ...     "user_id": [0, 1, 1, 1, 2],
    ...     "item_id": [1, 3, 0, 2, 1],
    ...     "score": [1, 0, 1, 1, 0]
    ... })
    >>> triplet_df = ItemSpecificSampler.rating2triplet(rating_matrix=rating_matrix, value_field="score")
    >>> triplet_df   # doctest: +NORMALIZE_WHITESPACE
             pos  neg
    item_id
    0        [1]   []
    1        [0]  [2]
    2        [1]   []
    3         []  [1]
    >>> sampler = UserSpecificPairSampler(triplet_df)
    >>> sampler([0, 1, 2], 5, pad_value=-1)
    [(0, [-1, -1, -1, -1, -1]), (1, [2, -1, -1, -1, -1]), (0, [-1, -1, -1, -1, -1])]
    >>> sampler([0, 1, 2], 5, neg=False, pad_value=-1)
    [(1, [1, -1, -1, -1, -1]), (1, [0, -1, -1, -1, -1]), (1, [1, -1, -1, -1, -1])]
    >>> sampler(rating_matrix["item_id"], 2, neg=rating_matrix["score"],
    ...     excluded_key=rating_matrix["user_id"], pad_value=-1)
    [(1, [2, -1]), (0, [-1, -1]), (0, [-1, -1]), (0, [-1, -1]), (1, [0, -1])]
    >>> sampler(rating_matrix["item_id"], 2, neg=rating_matrix["score"],
    ...     excluded_key=rating_matrix["user_id"], pad_value=-1, return_column=True)
    ((1, 0, 0, 0, 1), ([2, -1], [-1, -1], [-1, -1], [-1, -1], [0, -1]))
    >>> sampler(rating_matrix["item_id"], 2, neg=rating_matrix["score"],
    ...     excluded_key=rating_matrix["user_id"], pad_value=-1, return_column=True, split_sample_to_column=True)
    ((1, 0, 0, 0, 1), [(2, -1, -1, -1, 0), (-1, -1, -1, -1, -1)])
    """

    def __init__(self, triplet_df: pd.DataFrame,
                 query_field="item_id", pos_field="pos", neg_field="neg", set_index=False,
                 item_id_range=None, user_id_range=None,
                 random_state=10):
        super(ItemSpecificSampler, self).__init__(
            triplet_df=triplet_df,
            query_field=query_field,
            pos_field=pos_field,
            neg_field=neg_field,
            set_index=set_index,
            query_range=item_id_range,
            key_range=user_id_range,
            random_state=random_state)

    @staticmethod
    def rating2triplet(rating_matrix: pd.DataFrame,
                       query_field="item_id", key_field="user_id", value_field=None,
                       value_threshold=None, pos_field="pos", neg_field="neg",
                       *args, verbose=True, **kwargs):
        return TripletPairSampler.rating2triplet(
            rating_matrix=rating_matrix,
            query_field=query_field,
            key_field=key_field,
            value_field=value_field,
            value_threshold=value_threshold,
            pos_field=pos_field,
            neg_field=neg_field,
            *args,
            verbose=verbose,
            **kwargs
        )
